import numpy as np
import matplotlib.pyplot as plt
import scipy.signal as signal

#---------------------------------------------------------------------------------
#                                       Processing
#---------------------------------------------------------------------------------

def calculate_spectrum(rcv, dt):
    """
    Perform frequency spectrum analysis of seismic data and return the spectrum data.
    
    Parameters:
    rcv (np.ndarray): Seismic data with shape (n_receivers, n_samples).
    t (np.ndarray): Time vector with shape (n_samples,).
    
    Returns:
    freqs (np.ndarray): Array of frequencies (positive frequencies).
    amplitude_spectrum (np.ndarray): Amplitude spectrum of the seismic data.
    power_spectrum (np.ndarray): Power spectrum of the seismic data.
    """
    # 1. Perform Fast Fourier Transform (FFT) on each receiver's data
    # rcv.shape = (n_receivers, n_samples)
    n_receivers, n_samples = rcv.shape
    fs = 1 / dt  # Sampling frequency
    
    # Apply FFT to each trace (receiver data)
    fft_data = np.fft.fft(rcv, axis=1)  # FFT along the time axis (axis=1)
    
    # Frequency axis: Frequency bins corresponding to FFT result
    freqs = np.fft.fftfreq(n_samples, dt)  # Frequency bins
    
    # Only keep positive frequencies (real-valued spectrum)
    positive_freqs = freqs[:n_samples // 2]
    fft_data = fft_data[:, :n_samples // 2]
    
    # Compute amplitude spectrum
    amplitude_spectrum = np.abs(fft_data)
    
    # Compute power spectrum (optional)
    power_spectrum = amplitude_spectrum**2
    
    return positive_freqs, amplitude_spectrum, power_spectrum

def filter_low_frequencies_zero_phase(data, dt, cutoff_freq=5):
    """
    Apply zero-phase filtering to remove frequencies below the cutoff frequency.
    
    Parameters:
    data (np.ndarray): Seismic data with shape [shot, t, rcv], where
                        - shot: number of shots
                        - t: number of time samples
                        - rcv: number of receivers
    dt (float): Time step (sampling interval).
    cutoff_freq (float): The cutoff frequency in Hz. Frequencies below this value will be filtered out.
    
    Returns:
    np.ndarray: Zero-phase filtered seismic data.
    """
    n_shots, n_time, n_receivers = data.shape
    
    # Define the Nyquist frequency
    nyquist_freq = 0.5 / dt
    
    # Create a lowpass filter using scipy
    # The filter is designed to remove frequencies below cutoff_freq (lowpass filter)
    nyquist = 0.5 / dt  # Nyquist frequency
    normalized_cutoff = cutoff_freq / nyquist  # Normalize the cutoff frequency
    
    # Design a Butterworth lowpass filter
    b, a = signal.butter(6, normalized_cutoff, btype='high')
    
    # Apply zero-phase filtering to each trace using filtfilt
    filtered_data = np.zeros_like(data)
    for i in range(n_shots):
        for j in range(n_receivers):
            filtered_data[i, :, j] = signal.filtfilt(b, a, data[i, :, j])
    
    return filtered_data

#---------------------------------------------------------------------------------
#                                       Figuring
#---------------------------------------------------------------------------------

def plot_spectrum(positive_freqs, amplitude_spectrum, power_spectrum,
                  y1lim = [],y2lim=[],
                  save_path="",show=True):
    """
    Plot the frequency spectrum (amplitude and power) of seismic data.
    
    Parameters:
    positive_freqs (np.ndarray): Array of positive frequencies.
    amplitude_spectrum (np.ndarray): Amplitude spectrum of the seismic data.
    power_spectrum (np.ndarray): Power spectrum of the seismic data.
    
    Returns:
    None
    """
    # 2. Plot the frequency spectrum for one receiver (for example)
    plt.figure(figsize=(12, 4))
    plt.subplot(121)
    plt.plot(positive_freqs, amplitude_spectrum[0], label="Receiver 1 Amplitude Spectrum", color='blue')
    plt.title("Frequency Spectrum of Seismic Data")
    plt.xlabel("Frequency (Hz)")
    plt.ylabel("Amplitude")
    plt.xscale('log')
    plt.yscale('log')
    if len(y1lim) == 2:
        plt.ylim((y1lim[0],y1lim[1]))
    plt.grid(True)
    plt.legend()
    
    # 3. Plot the power spectrum (optional)
    plt.subplot(122)
    plt.plot(positive_freqs, power_spectrum[0], label="Receiver 1 Power Spectrum", color='red')
    plt.title("Power Spectrum of Seismic Data")
    plt.xlabel("Frequency (Hz)")
    plt.ylabel("Power")
    plt.xscale('log')
    plt.yscale('log')
    if len(y2lim) == 2:
        plt.ylim((y2lim[0],y2lim[1]))
    plt.grid(True)
    plt.legend()
    if not save_path == "":
        plt.savefig(save_path,bbox_inches="tight")
    if show:
        plt.show()

def plot_frequency_distribution(positive_freqs, amplitude_spectrum,
                                xlim = [],
                                save_path="",show=True):
    """
    Plot the frequency distribution of seismic data across all receivers.
    
    Parameters:
    positive_freqs (np.ndarray): Array of positive frequencies.
    amplitude_spectrum (np.ndarray): Amplitude spectrum of the seismic data, shape (n_receivers, n_freq_bins).
    
    Returns:
    None
    """
    # Sum the amplitude spectrum across all receivers to get the total distribution
    total_amplitude = np.sum(amplitude_spectrum, axis=0)
    
    # Plot the frequency distribution
    plt.figure(figsize=(10, 6))
    plt.plot(positive_freqs, total_amplitude, label="Total Amplitude Spectrum", color='green')
    plt.title("Frequency Distribution of Seismic Data")
    plt.xlabel("Frequency (Hz)")
    plt.ylabel("Total Amplitude")
    plt.grid(True)
    plt.legend()
    if len(xlim) == 2:
        plt.xlim((xlim[0],xlim[1]))
    if not save_path == "":
        plt.savefig(save_path,bbox_inches="tight")
    if show:
        plt.show()

def plot_filtered_data(original_data, filtered_data, shot_idx=0,cutoff_freq=5,save_path="",show=True):
    """
    Plot original and filtered seismic data for a specific shot.
    
    Parameters:
    original_data (np.ndarray): Original seismic data.
    filtered_data (np.ndarray): Filtered seismic data.
    shot_idx (int): Index of the shot to be plotted (default is 0).
    
    Returns:
    None
    """
    plt.figure(figsize=(10, 6))
    plt.subplot(1, 2, 1)
    plt.imshow(original_data[shot_idx], label='Original Data',aspect='auto',cmap='gray')
    plt.title(f"Original",fontsize=13)
    plt.ylabel("Time (samples)",fontsize=13)
    plt.xlabel("Receiver ID",fontsize=13)
    plt.grid(True,linestyle='--',color="silver",alpha=0.5)
    
    plt.subplot(1, 2, 2)
    plt.imshow(filtered_data[shot_idx], label='Filtered Data', aspect='auto',cmap='gray')
    plt.title(f"Filtered (below {cutoff_freq}Hz removed)",fontsize=13)
    plt.ylabel("Time (samples)",fontsize=13)
    plt.xlabel("Receiver ID",fontsize=13)
    plt.grid(True,linestyle='--',color="silver",alpha=0.5)
    plt.tight_layout()
    
    if not save_path == "":
        plt.savefig(save_path,bbox_inches="tight")
    if show:
        plt.show()

# # Example usage:
# # Assuming `data` is your seismic data with shape [shot, t, rcv], and `dt` is the time step
# cutoff_freq = 9
# filtered_data = filter_low_frequencies_zero_phase(data, dt, cutoff_freq=cutoff_freq)

# # Plot original and filtered data for a specific shot (e.g., shot 0)
# plot_filtered_data(data, filtered_data, shot_idx=0, cutoff_freq=cutoff_freq)


'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-04-20 09:32:43
* LastEditors: LiuFeng
* LastEditTime: 2024-05-15 19:30:38
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''

from .base import Misfit
import torch
import torch.nn as nn


class Misfit_waveform_smoothL1(Misfit): 
    """smoothed L1 misfit function (Ross Girshick, 2015, International conference on computer vision)
        https://pytorch.org/docs/stable/generated/torch.nn.SmoothL1Loss.html
        
    Paraemters:
    -----------
        obs (Tensors)   : the observed waveform 
        syn (Tensors)   : the synthetic waveform 
    """
    def __init__(self,dt = 1) -> None:
        self.dt = dt
        super().__init__()
    
    def forward(self,obs,syn):
        loss_fun = nn.SmoothL1Loss(reduction='sum',beta=1.0)
        loss     = loss_fun(obs,syn)*self.dt 
        return loss


'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-04-17 21:41:11
* LastEditors: LiuFeng
* LastEditTime: 2024-05-05 14:22:45
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''
import numpy as np
from math import log
import torch


# ADFWI DAMP
def bc_pml(nx,nz,dx,dz,pml,vmax,free_surface=True):
    """
        calculate the damping in both x and z direction
    """
    if free_surface:
        nx_pml = nx + 2*pml
        nz_pml = nz + pml
    else:
        nx_pml = nx + 2*pml
        nz_pml = nz + 2*pml
    damp_global = np.zeros((nx_pml,nz_pml))
    damp        = np.zeros(pml)
    a = (pml-1)*dx
    # Adjust the damping effect.
    R = 1e-6
    # R = 1e-3
    kappa = -3.0*vmax*log(R)/(2.0*a)
    for ix in range(0,pml):
        xa = ix*dx/a
        damp[ix] = kappa*xa*xa
        
    for ix in range(0,pml):
        for iz in range(0,nz_pml):
            damp_global[pml-ix-1,iz]            = damp[ix]
            damp_global[nx_pml+ix-pml,iz]       = damp[ix]

    for iz in range(0,pml):
        for ix in range((pml-(iz-1))-1,nx_pml-(pml-(iz))):
            if not free_surface:
                damp_global[ix,pml-iz-1]        = damp[iz]
            damp_global[ix,nz_pml+iz-pml]       = damp[iz]
            
    return damp_global.T

# SinCos damp
def bc_sincos(nx,nz,dx,dz,pml,free_surface=False):
    ''' Set up damping profile
    '''
    if free_surface:
        nx_pml = nx + 2*pml
        nz_pml = nz +   pml
    else:
        nx_pml = nx + 2*pml
        nz_pml = nz + 2*pml
    damp = np.ones((nz_pml, nx_pml))

    for i in range(pml):
        if not free_surface:
            damp[i, :]      *=  np.sin(np.pi/2 * i/pml)**2
        damp[-i-1, :]   *=  np.sin(np.pi/2 * i/pml)**2
        damp[:, i]      *=  np.sin(np.pi/2 * i/pml)**2
        damp[:, -i-1]   *=  np.sin(np.pi/2 * i/pml)**2

    return damp

# FDWave3D damp
def bc_gerjan(nx,nz,dx,dz,pml,alpha=0.0053,free_surface=True):
    """ PML Gerjan et al., 1985
        G = exp(a*[I - i]^2)
        -a: attenuation factor
    """
    wt = np.exp(-(alpha*(pml-np.arange(1,pml+1)))**2)
    if free_surface:
        nx_pml = nx + 2*pml
        nz_pml = nz + pml
    else:
        nx_pml = nx + 2*pml
        nz_pml = nz + 2*pml
    damp = np.ones((nz_pml,nx_pml))
    if free_surface:
        for k in range(1,len(wt)+1):
            damp[:nz_pml-k+1,k-1]               = wt[k-1]        # left
            damp[nz_pml-k   ,k-1:nx_pml-k+1]    = wt[k-1]        # bottom
            damp[:nz_pml-k+1,nx_pml-k]          = wt[k-1]        # right
    else:
        for k in range(1,len(wt)+1):
            damp[k-1:nz_pml-k+1 ,k-1]            = wt[k-1]        # left
            damp[k:nz_pml-k+1   ,nx_pml-k]       = wt[k-1]        # right
            damp[nz_pml-k       ,k-1:nx_pml-k+1] = wt[k-1]        # bottom
            damp[k-1            ,k-1:nx_pml-k+1] = wt[k-1]        # top
    return damp

# FDWave3D ABCdamp
def bc_pml_xz(nx,nz,dx,dz,pml,vmax,free_surface=True):
    if free_surface:
        nx_pml = nx + 2*pml
        nz_pml = nz + pml
    else:
        nx_pml = nx + 2*pml
        nz_pml = nz + 2*pml
    R = 1e-6
    ppml = -np.log(R)*3*vmax/(2*pml**3)
    BCx = np.zeros((nz_pml,nx_pml))
    BCz = np.zeros((nz_pml,nx_pml))
    if free_surface:
        for k in range(1,pml+1):
            BCx[:,k-1]           = (pml-k+1)**2*ppml/dx
            BCx[:,nx_pml-k]      = (pml-k+1)**2*ppml/dx
            BCz[nz_pml-k,:]      = (pml-k+1)**2*ppml/dz
    else:
        for k in range(1,pml+1):
            BCx[:,k-1]           = (pml-k+1)**2*ppml/dx
            BCx[:,nx_pml-k]      = (pml-k+1)**2*ppml/dx
            BCz[k-1,:]           = (pml-k+1)**2*ppml/dz
            BCz[nz_pml-k,:]      = (pml-k+1)**2*ppml/dz
    return BCx,BCz


"""
* Modified from: https://github.com/seisfwi/SWIT
* Original Author: Haipeng Li
* Original Author Email: haipeng@stanford.edu
=========================================================
* Author: Liu Feng (SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-04-20 09:32:43
* LastEditors: Liu Feng
* LastEditTime: 2024-05-15 19:30:38
* Description: 
* Copyright (c) 2024 by Liu Feng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
"""

from typing import List
import numpy as np
from ADFWI.utils import list2numpy,numpy2list

class Receiver(object):
    """Seismic Receiver class
    Parameters:
    -------------
    nt (int)    : number of time samples in the receiver data
    dt (float)  : Time inverval of data
    
    Notes:  1. The seismic data is assumed to start at time 0, e.g., ot = 0.
            2. The receiver locations should be added using the add_receiver method.
            3. The receiver is assumed to be the same for all shots, i.e., the
                receiver locations are not shot-dependent. This consideration is
                based on the fact the regular data (nshot, nrec, nt) can be dealt
                more efficiently on GPUs than irregular data. However, this may 
                limit the flexibility of modeling streamer data where the receiver
                locations are shot-dependent. The workaround solution is to apply 
                offset masking to the regular data to mimic the streamer data.
    """
    def __init__(self,nt:int,dt:float) -> None:
        self.nt = nt
        self.dt = dt
        
        self.loc_x  = []
        self.loc_z  = []
        self.locs   = []
        self.type   = []
        self.num    = 0
    
    def __repr__(self):
        """Print the receiver information
        """
        try:
            rcv_x = np.array(self.loc_x)
            rcv_z = np.array(self.loc_z)
            xmin = rcv_x.min()
            xmax = rcv_x.max()
            zmin = rcv_z.min()
            zmax = rcv_z.max()

            info = f"Seismic Receiver:\n"
            info += (
                f"  Receiver data   : {self.nt} samples at {self.dt * 1000:.2f} ms\n"
            )
            info += f"  Receiver number : {self.num}\n"
            info += f"  Receiver types  : {self.get_type(unique = True)}\n"
            info += f"  Receiver x range: {xmin} - {xmax} (grids)\n"
            info += f"  Receiver z range: {zmin} - {zmax} (grids)\n"
        except:
            info = f"Seismic Receiver:\n"
            info += f"  empty\n"

        return info
    
    def add_receivers(self, rcv_x: np.array,rcv_z:np.array, rcv_type: str) -> None:
        """add multiple receiver with same type
        """
        if rcv_x.shape != rcv_z.shape:
            raise ValueError(
                "Receiver Error: Inconsistant number of receiver in X and Z directions"
            )
        if rcv_type.lower() not in ["pr", "vx", "vz"]:
            raise ValueError("Receiver type must be either pr, vx, vz")
        rcv_n = len(rcv_x)
        # add the receiver
        self.loc_x.extend(numpy2list(rcv_x.reshape(-1)))
        self.loc_z.extend(numpy2list(rcv_z.reshape(-1)))
        self.type.extend([rcv_type]*rcv_n)
        self.num += rcv_n
    
    def add_receiver(self, rcv_x:int,rcv_z:int, rcv_type: str) -> None:
        """Append single receiver
        """
        if rcv_type.lower() not in ["pr", "vx", "vz"]:
            raise ValueError("Receiver type must be either pr, vx, vz")
        # add the receiver
        self.loc_x.append(rcv_x)
        self.loc_z.append(rcv_z)
        self.type.append(rcv_type)
        self.num += 1
    
    def get_loc(self):
        """Return the source location
        """
        rcv_x = list2numpy(self.loc_x).reshape(-1,1)
        rcv_z = list2numpy(self.loc_z).reshape(-1,1)
        rcv_loc = np.hstack((rcv_x,rcv_z))
        self.loc = rcv_loc.copy()
        return rcv_loc 
    
    def get_type(self, unique=False) -> List[str]:
        """Return the source type
        """
        type = list2numpy(self.type)
        
        if unique:
            type = list2numpy(list(set(self.type)))
        return type


'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-05-03 22:33:52
* LastEditors: LiuFeng
* LastEditTime: 2024-05-15 18:41:40
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''
from typing import Optional
import numpy as np
import torch
import matplotlib.pyplot as plt
from ADFWI.model import AbstractModel
from ADFWI.survey import Survey
from .boundary_condition import bc_pml_xz,bc_gerjan,bc_sincos
from .elastic_kernels import forward_kernel
from ADFWI.utils import numpy2tensor
import torch.jit as jit

class ElasticPropagator(torch.nn.Module):
# class ElasticPropagator(jit.ScriptModule):
    """The class of defining the propagator for the isotropic elastic wave
    equation (stress-velocity form), which is solved by the finite
    difference method.

    Parameters:
    -----------
    model: IsotropicElasticModel
        The model object
    survey: Survey
        The survey object
    """
    def __init__(self,
                model:AbstractModel,
                survey:Survey,
                device:Optional[str]    = 'cpu',
                cpu_num:Optional[int]   = 1,
                gpu_num:Optional[int]   = 1,
                dtype                   = torch.float32
                ):
        super().__init__()
        
        if not isinstance(model, AbstractModel):
            raise ValueError("model is not AbstractModel")

        if not isinstance(survey, Survey):
            raise ValueError("survey is not Survey")
        
        # ---------------------------------------------------------------
        # set the model and survey
        # ---------------------------------------------------------------
        self.model          = model
        self.survey         = survey
        self.device         = device
        self.dtype          = dtype
        self.cpu_num        = cpu_num
        self.gpu_num        = gpu_num
        
        # ---------------------------------------------------------------
        # parse parameters for model
        # ---------------------------------------------------------------
        self.ox, self.oz    = model.ox,model.oz
        self.dx, self.dz    = model.dx,model.dz
        self.nx, self.nz    = model.nx,model.nz
        self.nt             = survey.source.nt
        self.dt             = survey.source.dt
        self.f0             = survey.source.f0
        
        # ---------------------------------------------------------------
        # set the boundary: [top, bottom, left, right]
        # ---------------------------------------------------------------
        self.abc_type       = model.abc_type
        self.nabc           = model.nabc
        self.free_surface   = model.free_surface
        self.bcx,self.bcz,self.damp   = None,None,None
        self.boundary_condition()
        
        # ---------------------------------------------------------------
        # parameters for source
        # ---------------------------------------------------------------
        self.source         = self.survey.source
        self.src_loc        = self.source.get_loc()
        self.src_x          = numpy2tensor(self.src_loc[:,0],torch.long).to(self.device)
        self.src_z          = numpy2tensor(self.src_loc[:,1],torch.long).to(self.device)
        self.src_n          = self.source.num
        self.wavelet        = numpy2tensor(self.source.get_wavelet(),self.dtype).to(self.device)
        self.moment_tensor  = numpy2tensor(self.source.get_moment_tensor(),self.dtype).to(self.device)
        
        # ---------------------------------------------------------------
        # parameters for receiver
        # ---------------------------------------------------------------
        self.receiver       = self.survey.receiver
        self.rcv_loc        = self.receiver.get_loc()
        self.rcv_x          = numpy2tensor(self.rcv_loc[:,0],torch.long).to(self.device)
        self.rcv_z          = numpy2tensor(self.rcv_loc[:,1],torch.long).to(self.device)
        self.rcv_n          = self.receiver.num
        
        self.receiver_masks     = self.survey.receiver_masks
        self.receiver_masks_obs = self.survey.receiver_masks_obs
        
    def boundary_condition(self):
        if self.abc_type == "PML":
            bcx,bcz = bc_pml_xz(self.nx,self.nz,self.dx,self.dz,pml=self.nabc,
                                vmax=self.model.vp.cpu().detach().numpy().max(),
                                free_surface=self.free_surface)
            self.bcx = numpy2tensor(bcx,self.dtype).to(self.device)
            self.bcz = numpy2tensor(bcz,self.dtype).to(self.device)
        elif self.abc_type=='gerjan':
            damp = bc_gerjan(self.nx,self.nz,self.dx,self.dz,pml=self.nabc,alpha=self.model.abc_jerjan_alpha,
                            free_surface=self.free_surface)
            self.damp = numpy2tensor(damp,self.dtype).to(self.device)
        else:
            damp = bc_sincos(self.nx,self.nz,self.dx,self.dz,pml=self.nabc,
                             free_surface=self.free_surface)
            self.damp = numpy2tensor(damp,self.dtype).to(self.device)
        return 
    
    def forward(self,
                model:Optional[AbstractModel] = None,
                shot_index=None,
                fd_order = 4,
                checkpoint_segments=1):
        # calculate the thomson/lame and elastic moduli parameters
        model = self.model if model is None else model
        model.forward()
        
        # foward simulation for select shots
        src_x = self.src_x[shot_index] if shot_index is not None else self.src_x
        src_z = self.src_z[shot_index] if shot_index is not None else self.src_z
        src_n = len(src_x)
        wavelet = self.wavelet[shot_index] if shot_index is not None else self.wavelet
        moment_tensor = self.moment_tensor[shot_index] if shot_index is not None else self.moment_tensor
        
        record_waveform = forward_kernel(
            self.nx,self.nz,self.dx,self.dz,self.nt,self.dt,
            self.nabc,self.free_surface,
            src_x,src_z,src_n,wavelet,moment_tensor,
            self.rcv_x,self.rcv_z,self.rcv_n,
            self.abc_type,self.bcx,self.bcz,self.damp,
            model.lamu,model.lam,model.bx,model.bz,model.CC,
            fd_order=fd_order,n_segments=checkpoint_segments,
            device=self.device,dtype=self.dtype
        )
        return record_waveform


'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-04-26 19:42:24
* LastEditors: LiuFeng
* LastEditTime: 2024-05-13 23:16:49
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''

from ADFWI.model import AbstractModel
from ADFWI.utils import numpy2tensor
from ADFWI.model.parameters import (thomsen_init,elastic_moduli_init,
                         vs_vp_to_Lame,thomsen_to_elastic_moduli,
                         elastic_moduli_for_isotropic,elastic_moduli_for_TI,
                         parameter_staggered_grid)
from ADFWI.view import (plot_vp_vs_rho,plot_eps_delta_gamma,plot_lam_mu,plot_model)
from typing import Optional,Tuple,Union
import torch
from torch import Tensor
import numpy as np
from torchinfo import summary

class DIP_VTIModel(AbstractModel):
    """Elastic Velocity model with parameterization of vp and rho
    Parameters:
    --------------
        ox (float),oz(float)        : Origin of the model in x- and z- direction (m)
        nx (int),nz(int)            : NUmber of grid points in x- and z- direction (m)
        dx (float),dz(float)        : Grid size in x- and z- direction (m)
        vp_bound (tuple,Optional)   : Bounds for the P-wave velocity model, default None
        rho_bound (tuple,Optional)  : Bounds for the density model, default None
        vp_grad (bool,Optional)     : Flag for gradient of P-wave velocity model, default is False
        rho_grad (bool,Optional)    : Flag for gradient of density, default is False
        free_surface (bool,Optional): Flag for free surface, default is False
        abc_type (str)              : the type of absorbing boundary conditoin : PML,jerjan and other
        abc_jerjan_alpha (float)    : the attenuation factor for jerjan boundary condition
        nabc (int)                  : Number of absorbing boundary cells, default is 20
        device (str,Optional)       : The runing device
        dtype (dtypes,Optional)     : The dtypes for pytorch variable, default is torch.float32
    """
    def __init__(self,
                ox:float,oz:float,
                nx:int  ,nz:int,
                dx:float,dz:float,
                DIP_model_vp                                     = None,     # deep image prior models
                DIP_model_vs                                     = None,     
                DIP_model_rho                                    = None,
                DIP_model_eps                                    = None,
                DIP_model_delta                                  = None,
                DIP_model_gamma                                  = None,
                vp_bound:Optional[Tuple[float,float]]            = None,     # model parameter's boundary
                vs_bound:Optional[Tuple[float,float]]            = None,     
                rho_bound:Optional[Tuple[float,float]]           = None,
                eps_bound: Optional[Tuple[float, float]]         = None,
                delta_bound: Optional[Tuple[float, float]]       = None,
                gamma_bound: Optional[Tuple[float, float]]       = None,
                vp_init:Optional[Union[np.array,Tensor]]         = None,     # initial model parameter
                vs_init:Optional[Union[np.array,Tensor]]         = None,     
                rho_init:Optional[Union[np.array,Tensor]]        = None,
                eps_init:Optional[Union[np.array,Tensor]]        = None,
                delta_init:Optional[Union[np.array,Tensor]]      = None,
                gamma_init:Optional[Union[np.array,Tensor]]      = None,
                water_layer_mask:Optional[Union[np.array,Tensor]]= None,    # mask
                free_surface:Optional[bool]                      = False,
                anisotropic_type:Optional[str]                   = "vti",
                abc_type:Optional[str]                           = 'PML',
                abc_jerjan_alpha:Optional[float]                 = 0.0053,
                nabc:Optional[int]                               = 20,
                auto_update_rho:Optional[bool]                   = False,    # auto update parameters
                auto_update_vp:Optional[bool]                    = False,
                device                                           = 'cpu',
                dtype                                            = torch.float32
                )->None:
        # initialize the common model parameters
        super().__init__(ox,oz,nx,nz,dx,dz,free_surface=free_surface,abc_type=abc_type,abc_jerjan_alpha=abc_jerjan_alpha,nabc=nabc,device=device,dtype=dtype)
        self.anisotropic_type = anisotropic_type
        
        # update rho/vp using the empirical function
        self.auto_update_rho = auto_update_rho
        self.auto_update_vp  = auto_update_vp
        
        # gradient mask
        if water_layer_mask is not None:
            self.water_layer_mask = numpy2tensor(water_layer_mask,dtype=torch.bool).to(device)
        else:
            self.water_layer_mask = None
        
        # Neural networks
        self.pars           = ["vp","vs","rho","eps","gamma","delta"]
        self.DIP_model_vp   = DIP_model_vp
        self.DIP_model_vs   = DIP_model_vs
        self.DIP_model_rho  = DIP_model_rho
        self.DIP_model_eps  = DIP_model_eps
        self.DIP_model_delta= DIP_model_delta
        self.DIP_model_gamma= DIP_model_gamma
        if vp_init is not None:
            self.vp_init    = numpy2tensor(vp_init,dtype=dtype).to(device)
        if vs_init is not None:
            self.vs_init    = numpy2tensor(vs_init,dtype=dtype).to(device)
        if rho_init is not None:
            self.rho_init   = numpy2tensor(rho_init,dtype=dtype).to(device)
        if eps_init is not None:
            self.eps_init   = numpy2tensor(eps_init,dtype=dtype).to(device)
        if delta_init is not None:
            self.delta_init   = numpy2tensor(delta_init,dtype=dtype).to(device)
        if gamma_init is not None:
            self.gamma_init   = numpy2tensor(gamma_init,dtype=dtype).to(device)
        self.vp    = torch.zeros((nz,nx),dtype=dtype).to(device) if  vp_init   is None else self.vp_init.clone()
        self.vs    = torch.zeros((nz,nx),dtype=dtype).to(device) if  vs_init   is None else self.vs_init.clone()
        self.rho   = torch.zeros((nz,nx),dtype=dtype).to(device) if rho_init   is None else self.rho_init.clone()
        self.eps   = torch.zeros((nz,nx),dtype=dtype).to(device) if eps_init   is None else self.eps_init.clone()
        self.delta = torch.zeros((nz,nx),dtype=dtype).to(device) if delta_init is None else self.delta_init.clone()
        self.gamma = torch.zeros((nz,nx),dtype=dtype).to(device) if gamma_init is None else self.gamma_init.clone()
        self._parameterization_thomson()
        
        # initialize the lame constant
        self.mu         = None
        self.lamu       = None
        self.lam        = None
        self.muxz       = None
        self.b          = None
        self.bx         = None
        self.bz         = None
        self._parameterization_Lame()
        
        # initialize the elastic moduli
        self.CC         = []
        self._parameterization_elastic_moduli()

        # set model bounds
        self.lower_bound["vp"]      = vp_bound[0]    if vp_bound    is not None else None
        self.lower_bound["vs"]      = vs_bound[0]    if vs_bound    is not None else None
        self.lower_bound["rho"]     = rho_bound[0]   if rho_bound   is not None else None
        self.lower_bound['eps']     = eps_bound[0]   if eps_bound   is not None else None
        self.lower_bound['gamma']   = gamma_bound[0] if gamma_bound is not None else None
        self.lower_bound['delta']   = delta_bound[0] if delta_bound is not None else None
        self.upper_bound["vp"]      = vp_bound[1]    if vp_bound    is not None else None
        self.upper_bound["vs"]      = vs_bound[1]    if vs_bound    is not None else None
        self.upper_bound["rho"]     = rho_bound[1]   if rho_bound   is not None else None
        self.upper_bound["eps"]     = eps_bound[1]   if eps_bound   is not None else None
        self.upper_bound["gamma"]   = gamma_bound[1] if gamma_bound is not None else None
        self.upper_bound["delta"]   = delta_bound[1] if delta_bound is not None else None
        
        # check the input model
        self._check_bounds()
        self.check_dims()
        
    def get_requires_grad(self, par: str) -> bool:
        """Return the gradient of the model
        """
        if par not in self.pars:
            raise ValueError("Parameter {} not in model".format(par))
        if par == "vp":
            return self.DIP_model_vp is not None
        if par == "vs":
            return self.DIP_model_vs is not None
        if par == "rho":
            return self.DIP_model_rho is not None
        if par == "eps":
            return self.DIP_model_eps is not None
        if par == "delta":
            return self.DIP_model_delta is not None
        if par == "gamma":
            return self.DIP_model_gamma is not None

    def get_model(self, par: str):
        if par not in self.pars:
            raise "Error input parametrs"
        elif par == "vp":
            vp  = self.vp.cpu().detach().numpy()
            return vp
        elif par == "vs":
            vs  = self.vs.cpu().detach().numpy()
            return vs
        elif par == "rho":
            rho = self.rho.cpu().detach().numpy()
            return rho
        elif par == "eps":
            eps = self.eps.cpu().detach().numpy()
            return eps
        elif par == "delta":
            delta = self.delta.cpu().detach().numpy()
            return delta
        elif par == "gamma":
            gamma = self.gamma.cpu().detach().numpy()
            return gamma

    def get_bound(self, par: str) -> Tuple[float, float]:
        if par not in self.pars:
            raise "Error input parameters"
        else:
            m_min = self.lower_bound[par]
            m_max = self.upper_bound[par]        
        return [m_min,m_max]
    
    def __repr__(self) -> str:
        """Representation of the model object
        """
        info = f"   Model with parameters {self.pars}:\n"
        info += f"  Model orig: ox = {self.ox:6.2f}, oz = {self.oz:6.2f} m\n"
        info += f"  Model grid: dx = {self.dx:6.2f}, dz = {self.dz:6.2f} m\n"
        info += f"  Model dims: nx = {self.nx:6d}, nz = {self.nz:6d}\n"
        info += f"  Model size: {self.nx * self.nz * len(self.pars)}\n"
        info += f"  Free surface: {self.free_surface}\n"
        info += f"  Absorbing layers: {self.nabc}\n"
        info += f"  NN structure\n"
        if self.DIP_model_vp is not None:
            info += str(summary(self.DIP_model_vp,device=self.device))
        if self.DIP_model_vs is not None:
            info += str(summary(self.DIP_model_vs,device=self.device))
        if self.DIP_model_rho is not None:
            info += str(summary(self.DIP_model_rho,device=self.device))
        if self.DIP_model_eps is not None:
            info += str(summary(self.DIP_model_eps,device=self.device))
        if self.DIP_model_delta is not None:
            info += str(summary(self.DIP_model_delta,device=self.device))
        if self.DIP_model_gamma is not None:
            info += str(summary(self.DIP_model_gamma,device=self.device))
        return info
    
    def set_rho_using_empirical_function(self):
        """approximate rho via empirical relations with vp
        """
        vp          = self.vp.cpu().detach().numpy()
        rho         = self.rho.cpu().detach().numpy()
        rho_emprical= np.power(vp, 0.25) * 310
        if self.water_layer_mask is not None:
            grad_mask = self.water_layer_mask.cpu().detach().numpy()
            rho_emprical[grad_mask] = rho[grad_mask]
        self.rho    = numpy2tensor(rho_emprical,self.dtype).to(self.device)
        return
    
    def set_vp_using_empirical_function(self):
        """approximate vp via empirical relations with rho
        """
        rho         = self.rho.cpu().detach().numpy()
        vp          = self.vp.cpu().detach().numpy()
        vp_empirical= np.power(rho / 310, 4)
        if self.water_layer_mask is not None:
            grad_mask = self.water_layer_mask.cpu().detach().numpy()
            vp_empirical[grad_mask] = vp[grad_mask]
        self.vp     = numpy2tensor(vp_empirical,self.dtype).to(self.device)
        return   
    
    def _parameterization_thomson(self,*args,**kw_args):
        """setting variable and gradients
        """
        # get the model
        if self.DIP_model_vp is not None:
            self.vp = self.DIP_model_vp(*args,**kw_args)
        elif self.auto_update_vp:
            self.set_vp_using_empirical_function()
            
        if self.DIP_model_vs is not None:
            self.vs = self.DIP_model_vs(*args,**kw_args)
            
        if self.DIP_model_rho is not None:
            self.rho = self.DIP_model_rho(*args,**kw_args)
        elif self.auto_update_rho:
            self.set_rho_using_empirical_function()
            
        if self.DIP_model_eps is not None:
            self.eps = self.DIP_model_eps(*args,**kw_args)
        
        if self.DIP_model_delta is not None:
            self.delta = self.DIP_model_delta(*args,**kw_args)
        
        if self.DIP_model_gamma is not None:
            self.gamma = self.DIP_model_gamma(*args,**kw_args)
        return
    
    def _parameterization_Lame(self):
        """Calculate the lame parameters
        """
        mu,lamu,lam,b   = vs_vp_to_Lame(self.vp,self.vs,self.rho)
        self.mu         = mu
        self.lamu       = lamu
        self.lam        = lam
        self.b          = b
        return
    
    def _parameterization_elastic_moduli(self):
        """calculate the 21 dependent elastic moduli
        """
        # initialize elastic moduli
        CC = [C11,C12,C13,C14,C15,C16,C22,C23,C24,C25,C26,C33,C34,C35,C36,C44,C45,C46,C55,C56,C66] = elastic_moduli_init(self.nz,self.nx,self.device,self.dtype)
        # transform thomsen parameter to elastic moduli 
        C11,C13,C33,C44,C66 = thomsen_to_elastic_moduli(self.vp,self.vs,self.rho,self.eps,self.delta,self.gamma)
        CC = [C11,C12,C13,C14,C15,C16,C22,C23,C24,C25,C26,C33,C34,C35,C36,C44,C45,C46,C55,C56,C66]
        # define elastic moduli for isotropic model
        CC = [C11,C12,C13,C14,C15,C16,C22,C23,C24,C25,C26,C33,C34,C35,C36,C44,C45,C46,C55,C56,C66] = elastic_moduli_for_isotropic(CC)
        # prepare staggered grid settings
        bx,bz,muxz,C44,C55,C66 = parameter_staggered_grid(self.mu,self.b,C44,C55,C66,self.nx,self.nz)
        self.bx = bx
        self.bz = bz
        self.muxz = muxz
        CC = [C11,C12,C13,C14,C15,C16,C22,C23,C24,C25,C26,C33,C34,C35,C36,C44,C45,C46,C55,C56,C66]
        self.CC = CC
        return 
    
    def _plot_vp_vs_rho(self,**kwargs):
        """plot velocity model
        """
        plot_vp_vs_rho(self.vp,self.vs,self.rho,
                            dx=self.dx,dz=self.dz,**kwargs)
        return
    
    def _plot_eps_delta_gamma(self,**kwargs):
        """plot anisotropic parameters
        """
        plot_eps_delta_gamma(self.eps,self.delta,self.gamma,
                            dx=self.dx,dz=self.dz,**kwargs)
        return
    
    def _plot_lam_mu(self,**kwargs):
        """plot lame parameters
        """
        plot_lam_mu(self.lam,self.mu,
                            dx=self.dx,dz=self.dz,**kwargs)
        return
    
    def _plot(self,var,**kwargs):
        """plot single velocity model
        """
        model_data = self.get_model(var)
        plot_model(model_data,title=var,**kwargs)
        return
    
    def clip_params(self,par)->None:
        """Clip the model parameters to the given bounds
        """
        if self.get_requires_grad(par):
            if self.lower_bound[par] is not None and self.upper_bound[par] is not None:
                # Retrieve the model parameter
                m = getattr(self, par)
                min_value = self.lower_bound[par]
                max_value = self.upper_bound[par]
                # Create a temporary copy for masking purposes
                m_temp = m.clone()  # Use .clone() instead of .copy() to avoid issues with gradients

                # Clip the values of the parameter using in-place modification with .data
                m.data.clamp_(min_value, max_value)
                
                # Apply the water layer mask if it is not None, using in-place modification
                if self.water_layer_mask is not None:
                    m.data = torch.where(self.water_layer_mask.contiguous(), m_temp.data, m.data)
        return

    def forward(self,*args,**kwargs) -> Tuple:
        """Forward method of the elastic model class
        """
        self._parameterization_thomson()
            
        # Clip the model parameters
        self.clip_params("vp")
        self.clip_params("vs")
        self.clip_params("rho")
        self.clip_params("eps")
        self.clip_params("delta")
        self.clip_params("gamma")
        
        # calculate the thomson/lame and elastic moduli parameters
        self._parameterization_Lame()
        self._parameterization_elastic_moduli()
        return 


from setuptools import setup, find_packages

setup(
    name='ADFWI-Torch',
    version='0.1.2',
    packages=find_packages(),
    install_requires=[
        'numpy>=1.24.0',
        'scipy>=1.10.0',
        'matplotlib>=3.7.0',
        'torch>=2.0.0',
        'disba>=0.6.0',
        'torchinfo',
        'segyio',
        'ncg_optimizer',
        'pysdtw',
        'geomloss',
        'h5py',
        'obspy',
        'tqdm',
        'FyeldGenerator'
    ],
    author='Feng Liu',
    author_email='liufeng2317@sjtu.edu.cn',
    description='ADFWI: A framework for high-resolution subsurface parameter estimation using Full Waveform Inversion',
    long_description=open('README_PIP.md', encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/liufeng2317/ADFWI',
    license='MIT',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)



'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-03-01 20:51:32
* LastEditors: LiuFeng
* LastEditTime: 2024-03-21 10:49:51
* Description: 

* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''
import numpy as np
import torch 
import numpy as np
from scipy.signal import butter, filtfilt

##################################################################################
#                   multi-frequency processing
##################################################################################
# def lowpass(x1, highcut, fn, order=1, axis=1, show=False):
#     x = copy.deepcopy(x1)
#     # Zero padding
#     padding = 512
#     x = np.hstack((x, np.zeros((x.shape[0], padding, x.shape[2]))))
#     nt = x.shape[axis]
#     # Bring the data to frequency domain
#     x_fft = fft.fft(x, n=nt, axis=axis)
#     # Calculate the highcut btween 0 to 1
#     scaled_highcut = 2*highcut/fn
#     # Generate the filter
#     b, a = butter(order, scaled_highcut, btype='lowpass', output="ba")
#     # Get the frequency response
#     w, h1 = freqz(b, a, worN=nt, whole=True)
#     h = np.diag(h1)
#     # Apply the filter in the frequency domain
#     fd = h @ x_fft
#     #Double filtering by the conjugate to make up the shift
#     h = np.diag(np.conjugate(h1))
#     fd = h @ fd
#     # Bring back to time domaine
#     f_inv = fft.ifft(fd, n=nt, axis=axis).real
#     f_inv = f_inv[:, :-padding, :]
#     return f_inv

# def adj_lowpass(x, highcut, fn, order, axis=1):
#     # Zero padding
#     padding = 512
#     x = np.hstack((x, np.zeros((x.shape[0], padding, x.shape[2]))))
#     nt = x.shape[axis]
#     # Bring the data to frequency domain
#     x_fft = np.fft.fft(x, n=nt, axis=axis)
#     # Calculate the highcut btween 0 to 1
#     scaled_highcut = 2*highcut / fn
#     # Generate the filter
#     b, a = butter(order, scaled_highcut, btype='lowpass', output="ba")
#     # Get the frequency response
#     w, h = freqz(b, a, worN=nt, whole=True)
#     # Get the conjugate of the filter
#     h_c = np.diag(np.conjugate(h))
#     # Apply the adjoint filter in the frequency domain
#     fd = h_c @ x_fft
#     # Double filtering by the conjugate to make up the shift
#     h_c = np.diag(h)
#     fd = h_c @ fd
#     # Bring back to time domaine
#     adj_f_inv = np.fft.ifft(fd, axis=axis).real
#     adj_f_inv = adj_f_inv[:, :-padding, :]
#     return adj_f_inv

def lowpass(x, highcut, fn, order=1, axis=1):
    """
    Apply low-pass filter in the time domain using filtfilt (zero-phase filtering).
    
    Parameters:
    x (np.ndarray): Input signal (3D array: [nsrc, nt, nrcv]).
    highcut (float): High cutoff frequency in Hz.
    fn (float): Sampling frequency in Hz.
    order (int): Order of the Butterworth filter.
    
    Returns:
    np.ndarray: Low-pass filtered signal.
    """
    # Nyquist frequency
    nyquist = 0.5 * fn
    # Normalized cutoff frequency
    normal_cutoff = highcut / nyquist
    # Butterworth filter coefficients
    b, a = butter(order, normal_cutoff, btype='low', analog=False)
    
    # Apply the filter using filtfilt along the time axis (axis=1)
    # Apply the filter for all sources and receivers simultaneously (vectorized operation)
    y = np.empty_like(x)
    for i in range(x.shape[2]):  # Loop over receivers (nrcv)
        y[:, :, i] = filtfilt(b, a, x[:, :, i], axis=axis)
    
    return y


def adj_lowpass(x, highcut, fn, order=1, axis=1):
    """
    Apply adjoint low-pass filter in the time domain.
    
    Parameters:
    x (np.ndarray): Input signal (3D array: [nsrc, nt, nrcv]).
    highcut (float): High cutoff frequency in Hz.
    fn (float): Sampling frequency in Hz.
    order (int): Order of the Butterworth filter.
    
    Returns:
    np.ndarray: Adjoint low-pass filtered signal.
    """
    # Nyquist frequency
    nyquist = 0.5 * fn
    # Normalized cutoff frequency
    normal_cutoff = highcut / nyquist
    # Butterworth filter coefficients
    b, a = butter(order, normal_cutoff, btype='low', analog=False)
    
    # Reverse the time axis (nt), apply the filter, and reverse back (vectorized operation)
    adj_filtered = np.empty_like(x)
    for i in range(x.shape[2]):  # Loop over receivers (nrcv)
        # Apply the filter to the reversed signal and reverse it back
        adj_filtered[:, :, i] = np.flip(filtfilt(b, a, np.flip(x[:, :, i], axis=axis), axis=axis), axis=axis)
    return adj_filtered

def data2d_to_3d(data1_2d, data2_2d, ns, nr):
    nt = data1_2d.shape[0]
    
    data1_3d = torch.empty((ns, nt, nr))
    data2_3d = torch.empty((ns, nt, nr))
    
    for i in range(ns):
        data1_3d[i, :, :] = data1_2d[:, i*nr:(i+1)*nr]
        data2_3d[i, :, :] = data2_2d[:, i*nr:(i+1)*nr]
    return data1_3d, data2_3d


def data3d_to_2d(data1_3d, data2_3d):
    ns, nt, nr = data2_3d.shape
    x1_2d = torch.empty((nt, ns*nr))
    x2_2d = torch.empty((nt, ns*nr))
    for i in range(ns):
        x1_2d[:, i*nr:(i+1)*nr] = data1_3d[i, ...]
        x2_2d[:, i*nr:(i+1)*nr] = data2_3d[i, ...]
        
    return x1_2d, x2_2d

            
def lpass(x1, x2, highcut, fn):
    """
        fn is the sampling frequency
    """
    x1_filtered, x2_filtered = Lfilter.apply(x1, x2, highcut, fn)
    return x1_filtered, x2_filtered



class Lfilter(torch.autograd.Function):
    
    @staticmethod
    def forward(ctx, x1, x2, highcut, fn):
        ctx.lpass_highcut = highcut
        ctx.lpass_fn = fn
        
        ns, nt, nr = x1.shape
        device = x1.device
        
        x1,x2 = x1.detach(),x2.detach()
        x1,x2 = data3d_to_2d(x1 , x2)
        x1,x2 = torch.unsqueeze(x1, 0),torch.unsqueeze(x2, 0)
        
        filtered1 = lowpass(x1.numpy(), highcut=highcut, fn=fn, order=6, axis=1)
        filtered2 = lowpass(x2.numpy(), highcut=highcut, fn=fn, order=6, axis=1)

        filtered1, filtered2 = data2d_to_3d(
                torch.Tensor(filtered1[0, ...]),
                torch.Tensor(filtered2[0, ...]),
                ns, nr
            )
        filtered1 = torch.tensor(filtered1, device=device)
        filtered2 = torch.tensor(filtered2, device=device)
        return filtered1,filtered2
    
    @staticmethod
    def backward(ctx, adj1, adj2):
        
        ns, nt, nr = adj1.shape
        device = adj1.device.type
        
        x1,x2  = adj1.detach(),adj2.detach()
        x1, x2 = data3d_to_2d(x1, x2)
        x1,x2  = torch.unsqueeze(x1, 0),torch.unsqueeze(x2, 0)
        
        filtered1 = adj_lowpass(x1.numpy(), highcut=ctx.lpass_highcut, fn=ctx.lpass_fn, order=6, axis=1)
        
        filtered2 = adj_lowpass(x2.numpy(), highcut=ctx.lpass_highcut, fn=ctx.lpass_fn, order=6, axis=1)
        
        filtered1, filtered2 = data2d_to_3d(
            torch.Tensor(filtered1[0, ...]),
            torch.Tensor(filtered2[0, ...]),
            ns, nr)
        
        filtered1 = torch.tensor(filtered1, device=device)
        filtered2 = torch.tensor(filtered2, device=device)
                   
        return filtered1.to(device=device), \
                filtered2.to(device=device),\
                    None,\
                    None


'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-04-20 09:32:43
* LastEditors: LiuFeng
* LastEditTime: 2024-05-15 19:30:38
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''

from .base import Misfit
import torch
import numpy as np
import torch.nn.functional as F

@torch.jit.script
def hilbert(x: torch.Tensor) -> torch.Tensor:
    '''Perform Hilbert transform along the last axis of x.
    
    Args:
        x (Tensor): Input signal data.
        
    Returns:
        Tensor: Analytic signal with the same shape as `x`.
    '''
    device = x.device
    N = x.shape[-1] * 2  # Double the length for FFT
    Xf = torch.fft.fft(x, n=N)  # FFT on extended signal
    h = torch.zeros(N, dtype=Xf.dtype, device=device)  # Initialize multiplier array
    h[0] = 1
    h[1:(N + 1) // 2] = 2  # Set values to create analytic signal
    if N % 2 == 0:
        h[N // 2] = 1  # Special case for even-length signals
    return torch.fft.ifft(Xf * h)[..., :x.shape[-1]]  # Return Hilbert-transformed signal

@torch.jit.script
def diff(x: torch.Tensor, dim: int = -1, same_size: bool = False) -> torch.Tensor:
    '''Compute discrete difference along the last axis.
    
    Args:
        x (Tensor): Input tensor.
        dim (int): Axis along which to compute the difference (default: -1).
        same_size (bool): If True, pad the output to maintain the same size.
        
    Returns:
        Tensor: Discrete difference along specified axis.
    '''
    if same_size:
        return F.pad(x[..., 1:] - x[..., :-1], (1, 0))  # Pad to match original shape
    else:
        return x[..., 1:] - x[..., :-1]
    
@torch.jit.script
def unwrap(phi: torch.Tensor, dim: int = -1) -> torch.Tensor:
    '''Unwrap phase by correcting for phase discontinuities.
    
    Args:
        phi (Tensor): Phase tensor.
        dim (int): Axis along which to unwrap the phase (default: -1).
        
    Returns:
        Tensor: Unwrapped phase tensor.
    '''
    dphi = diff(phi, same_size=True)  # Calculate discrete phase difference
    dphi_m = ((dphi + np.pi) % (2 * np.pi)) - np.pi  # Map phase difference to [-pi, pi]
    dphi_m[(dphi_m == -np.pi) & (dphi > 0)] = np.pi  # Correct for edge cases
    phi_adj = dphi_m - dphi
    phi_adj[dphi.abs() < np.pi] = 0  # Adjust only where discontinuities exist
    return phi + phi_adj.cumsum(dim)  # Accumulate adjustments

class Misfit_envelope(Misfit):
    '''Compute the envelope misfit for initial velocity model estimation.
    
    References:
        Wu et al., 2014; Yuan et al., 2015
    
    Args:
        dt (float): Time sampling interval.
        p (float): Norm order for envelope difference.
        instaneous_phase (bool): If True, use instantaneous phase for misfit.
        norm (str): Norm type ("L1" or "L2") for final loss calculation.
    '''
    
    def __init__(self, dt: float = 1, p: float = 1.5, instaneous_phase: bool = False, norm: str = "L2") -> None:
        super().__init__()
        self.p = p
        self.instaneous_phase = instaneous_phase
        self.dt = dt
        self.norm = norm
        
    def forward(self, obs: torch.Tensor, syn: torch.Tensor) -> torch.Tensor:
        '''Compute the misfit between observed and synthetic waveforms.
        
        Args:
            obs (Tensor): Observed waveform [batch, trace, time].
            syn (Tensor): Synthetic waveform [batch, trace, time].
        
        Returns:
            Tensor: Envelope or phase difference loss.
        '''
        mask1    = torch.sum(torch.abs(obs),axis=1) == 0
        mask2    = torch.sum(torch.abs(syn),axis=1) == 0
        mask     = ~(mask1 * mask2)
        
        device = obs.device
        rsd = torch.zeros((obs.shape[0], obs.shape[2], obs.shape[1]), device=device)  # Residual storage
        
        for ishot in range(obs.shape[0]):
            trace_idx = torch.argwhere(mask[ishot]).reshape(-1)
            obs_shot = obs[ishot,:,trace_idx].squeeze(axis=0).T  # Transpose to [trace, time series]
            syn_shot = syn[ishot,:,trace_idx].squeeze(axis=0).T
            
            # Hilbert transform to get analytic signal
            analytic_signal_obs = hilbert(obs_shot)
            analytic_signal_syn = hilbert(syn_shot)
            
            # Compute envelopes (magnitude of analytic signals)
            envelopes_obs = torch.abs(analytic_signal_obs)
            envelopes_syn = torch.abs(analytic_signal_syn)
            
            if self.instaneous_phase:
                # Use instantaneous phase for misfit
                phase_obs = unwrap(torch.angle(analytic_signal_obs))
                phase_syn = unwrap(torch.angle(analytic_signal_syn))
                rsd[ishot,trace_idx,:] = (phase_obs - phase_syn).unsqueeze(0)
            else:
                # Compute envelope difference with norm p
                rsd[ishot,trace_idx,:] = (envelopes_syn**self.p - envelopes_obs**self.p).unsqueeze(0)

        # Compute final loss based on the selected norm
        if self.norm == "L1":
            loss = torch.sum(torch.abs(rsd))
        else:
            loss = 0.5 * torch.sum(rsd * rsd * self.dt)  # L2 loss with time weighting
        
        return loss


"""
* Modified from: https://github.com/seisfwi/SWIT
* Original Author: Haipeng Li
* Original Author Email: haipeng@stanford.edu
=========================================================
* Author: Liu Feng (SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-04-20 09:32:43
* LastEditors: Liu Feng
* LastEditTime: 2024-05-15 19:30:38
* Description: 
* Copyright (c) 2024 by Liu Feng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
"""

import numpy as np
import torch

#########################################################################################
#               optimizer from swit (https://github.com/seisfwi/SWIT)
#########################################################################################
class Optimization(object):
    '''
        Optimization based on Fortran subroutines from SEISCOPE OPTIMIZATION TOOLBOX
    '''
    def __init__(self, niter_max = 20, conv = 1e-8, method = 'SD', 
                 bound = False, lb = None, ub = None, 
                 lbfgs_memory = 20, debug = False):
        '''
            Initialize the optimizer class
        
            Parameters
            ----------
            niter_max (int)     : maximum number of iterations
            conv (float)        : convergence criterion
            method (str)        : optimization method, 'SD', 'CG', 'LBFGS', 'PLBFGS', 'TRN', 'PTRN'
            bound (bool)        : bound constraints, True or False
            lb (array)          : lower bound
            ub (array)          : upper bound
            lbfgs_memory (int)  : memory for LBFGS method
            debug (bool)        : debug mode, True or False
        '''
        self.niter_max = niter_max
        self.conv = conv
        self.method= method
        self.bound = bound
        self.lb = lb
        self.ub = ub
        self.debug = debug

        # Default parameters
        self.threshold = 100.           # Tolerance on bound constraints satisfaction
        self.FLAG = 'INIT'

        # Initialize linesearch parameters by default
        self.m1 = 0.0                   # Wolfe conditions parameter 1, 0.0 for no Wolfe conditions due to the scale of the misfit
        self.m2 = 0.95                  # Wolfe conditions parameter 2
        self.mult_factor = 5            # bracketting parameter
        self.nls_max = 10               # max number of linesearch
        self.cpt_ls = 0                 # linesearch counter
        self.first_ls = True            # first linesearch flag
        self.alpha = 0.02                 # first value for the linesearch steplength

        # Quasi-Newton l-BFGS method
        if self.method in ['LBFGS', 'PLBFGS']:
            self.l = lbfgs_memory 

        # Truncated Newton method
        if self.method in ['TRN', 'PTRN']:
            self.niter_max_CG=5         # maximum number of inner conjugate gradient iterations


    def iterate(self, x, fcost, grad, grad_preco):
        ''' perform nonlinear inversion
        '''

        # copy the input arrays and convert them to 1D arrays
        x           = np.copy(x.flatten())
        grad        = np.copy(grad.flatten())
        grad_preco  = np.copy(grad_preco.flatten())

        # Preconditioned Steepest Descent: PSTD
        if self.method == 'SD':
            x = self.PSTD(x, fcost, grad, grad_preco)

        # Preconditioned nonlinear conjugate gradient: PNLCG
        elif self.method == 'CG':
            x = self.PNLCG(x, fcost, grad, grad_preco)

        # Quasi-Newton l-BFGS method: LBFGS
        elif self.method == 'LBFGS':
            x = self.LBFGS(x, fcost, grad)

        # Preconditioned Quasi-Newton l-BFGS method: PLBFGS
        elif self.method == 'PLBFGS':
            x = self.PLBFGS(x, fcost, grad, grad_preco)
        
        # Truncated Newton method: TRN
        elif self.method == 'TRN':
            x = self.TRN(x, fcost, grad)

        # Preconditioned Truncated Newton method: TRN
        elif self.method == 'PTRN':
            x = self.PTRN(x, fcost, grad, grad_preco)

        else:
            raise AssertionError('method: SD, CG, LBFGS, PLBFGS, TRN, PTRN')

        return x


    #--------------------------------------------------------------------#
    #            Preconditioned Steepest Descent: PSTD                   # 
    #--------------------------------------------------------------------#
    def PSTD(self, x, fcost, grad, grad_preco):
        ''' Preconditioned Steepest Decent method
        '''
        # Initialize the linesearch process
        if(self.FLAG == 'INIT'):
            self.init_PSTD(x, fcost, grad, grad_preco)
            x = self.linesearch(x, fcost, grad)

            self.print_info(fcost)
            self.FLAG = 'GRAD'
            self.nfwd_pb += 1
        
        # call the linesearch process
        else:
            x = self.linesearch(x, fcost, grad)

            if(self.task == 'NEW_STEP'):
                self.cpt_iter += 1
                
                # test for convergence
                if(self.std_test_conv(fcost)):
                    self.FLAG = 'CONV'
                    self.print_info(fcost)
                
                # if a NEW_STEP is taken, compute a new descent direction using 
                # current descent, gradient and preconditioned gradient
                else:
                    self.FLAG = 'NSTE'
                    self.grad = grad
                    self.descent = -1. * grad_preco
                    self.print_info(fcost)

            # if the linesearch needs a new gradient then ask the user to provide it
            elif(self.task == 'NEW_GRAD'):
                self.FLAG = 'GRAD'
                self.nfwd_pb = self.nfwd_pb+1
            
            # if the linesearch has failed, inform the user
            elif(self.task == 'FAILURE!'):
                self.FLAG = 'FAIL'
                self.print_info(fcost)
        return x


    def init_PSTD(self, x, fcost, grad, grad_preco):
        '''  Initilize linesearch parameters
        '''

        # set counters
        self.cpt_iter = 0
        self.f0 = fcost
        self.nfwd_pb = 0

        # linesearch parameters
        self.fk = fcost

        # memory allocations
        self.xk = np.zeros_like(x)
        self.grad = np.zeros_like(x)
        self.descent = np.zeros_like(x)

        self.xk[:] = x[:]
        self.grad[:] = grad[:]

        # first descent direction
        self.descent[:] = -1. * grad_preco[:]


    #--------------------------------------------------------------------#
    #       Preconditioned nonlinear conjugate gradient: PNLCG           # 
    #--------------------------------------------------------------------#
    def PNLCG(self, x, fcost, grad, grad_preco):
        ''' Preconditioned nonlinear conjugate gradient
        '''
    
        # initialize the linesearch process
        if(self.FLAG == 'INIT'):

            self.init_PNLCG(x, fcost, grad, grad_preco)
            x = self.linesearch(x, fcost, grad)

            self.print_info(fcost)
            self.FLAG = 'GRAD'
            self.nfwd_pb += 1

            # Store current gradient before the user compute the new one
            self.grad_prev[:] = grad[:]
        
        # call the linesearch process
        else:
            x = self.linesearch(x, fcost, grad)

            if(self.task == 'NEW_STEP'): 
                self.cpt_iter += 1

                # test for convergence
                if(self.std_test_conv(fcost)):
                    self.FLAG = 'CONV'
                    self.grad[:] = grad[:]
                    self.print_info(fcost)
                
                # if a NEW_STEP is taken, compute a new descent direction using 
                # current descent, gradient and preconditioned gradient
                else:
                    self.FLAG = 'NSTE'
                    self.descent_PNLCG(grad, grad_preco)                    
                    self.grad[:] = grad[:]
                    self.print_info(fcost)

            # if the linesearch needs a new gradient then ask the user to provide it
            elif(self.task == 'NEW_GRAD'):
                self.FLAG = 'GRAD'
                self.nfwd_pb += 1
                
                # Store current gradient before the user compute the new one
                self.grad_prev[:] = grad[:]

            # if the linesearch has failed, inform the user
            elif(self.task == 'FAILURE!'):
                self.FLAG ='FAIL'
                self.grad[:] = grad[:]
                self.print_info(fcost)

        return x


    def init_PNLCG(self, x, fcost, grad, grad_preco):
        '''  Initilize linesearch parameters
        '''

        # set counters
        self.cpt_iter = 0
        self.f0 = fcost
        self.nfwd_pb = 0

        # linesearch parameters
        self.fk=fcost

        # memory allocations
        self.grad_prev = np.zeros_like(x)
        self.descent_prev = np.zeros_like(x)
        self.xk = np.zeros_like(x)
        self.grad = np.zeros_like(x)
        self.descent = np.zeros_like(x)

        self.xk[:] = x[:]
        self.grad[:] = grad[:]
            
        # first descent direction
        self.descent[:] = -1. * grad_preco[:]


    def descent_PNLCG(self, grad, grad_preco):
        ''' Descent direction
        '''

        # Storing old descent direction
        self.descent_prev[:] = self.descent[:]
        
        # The original PNLCG method (Dai and Yuan, 1999) in the SEISCOPE optimization toolbox
        # # Computation of beta (Dai and Yuan, 1999)
        # gkpgk = scalL2(grad, grad_preco)
        # skpk  = scalL2(grad - self.grad_prev, self.descent_prev)
        # if skpk != 0.:
        #     beta = gkpgk / skpk
        # else:
        #     beta = 0.

        # # Safeguard (may be useful in some cases)
        # if((beta >= 1e5) or (beta <= -1e5)):
        #     beta = 0.

        # Computation of beta (Polak and Ribière 1969), modified by Haipeng Li
        gkpgk = scalL2(grad, grad - self.grad_prev)
        skpk  = scalL2(self.grad_prev, self.grad_prev)
        if skpk != 0.:
            beta = gkpgk / skpk
        else:
            beta = 0.

        # Safeguard for beta (may be useful in some cases, empirical values)
        if beta > 2 or beta < 0:
            # print('Optimization: reset to descent direction in NLCG due to illegal beta = {:.2f}'.format(beta))
            beta = 0.
            
        # Computation of the descent direction
        self.descent[:] = -1. * grad_preco[:] + beta * self.descent_prev[:]
        

    #--------------------------------------------------------------------#
    #                Quasi-Newton l-BFGS method: LBFGS                   # 
    #--------------------------------------------------------------------#
    def LBFGS(self, x, fcost, grad):
        ''' Quasi-Newton l-BFGS method
        '''
    
        if(self.FLAG == 'INIT'):
            # initialize the linesearch process
            self.init_LBFGS(x, fcost, grad)
            x = self.linesearch(x, fcost, grad)
            self.print_info(fcost)
            self.FLAG = 'GRAD'
            self.nfwd_pb += 1
        else:
            # else call the linesearch process
            x = self.linesearch(x, fcost, grad)
            if(self.task == 'NEW_STEP'):
                self.cpt_iter += 1
                # test for convergence
                if(self.std_test_conv(fcost)):
                    self.FLAG = 'CONV'
                    self.grad[:] = grad[:]
                    self.print_info(fcost)
                else:
                    self.FLAG = 'NSTE'
                    # if a NEW_STEP is taken, compute a new descent direction using current gradient
                    # and l-BFGS approximation of the inverse Hessian preconditioned gradient

                    # LBFGS update        
                    self.update_LBFGS(x, grad)
                    
                    # Computation of the new descent direction
                    self.descent_LBFGS(grad)
                    
                    # LBFGS store
                    self.save_LBFGS(x, grad)
                    
                    # print info on current iteration
                    self.grad[:] = grad[:]
                    self.print_info(fcost)
                    
            elif(self.task == 'NEW_GRAD'):
                # if the linesearch needs a new gradient then ask the user to provide it
                self.FLAG = 'GRAD'
                self.nfwd_pb += 1

            elif(self.task == 'FAILURE!'):
                # if the linesearch has failed, inform the user       #
                self.FLAG='FAIL'
                # print info on current iteration
                self.grad[:] = grad[:]
                self.print_info(fcost)
        return x


    def init_LBFGS(self, x, fcost, grad):
        '''  Initilize linesearch parameters
        '''
        
        # set counters
        n = np.size(x)
        self.cpt_iter = 0
        self.f0 = fcost
        self.nfwd_pb = 0  
        self.sk = np.zeros((n, self.l))
        self.yk = np.zeros((n, self.l))
        self.cpt_lbfgs = 1
    
        # initialize linesearch parameters
        self.fk = fcost

        # memory allocations
        self.xk = np.zeros_like(x)
        self.grad = np.zeros_like(x)
        self.descent = np.zeros_like(x)

        self.xk[:] = x[:]
        self.grad[:] = grad[:]

        # first descent direction
        self.descent[:] = -1. * grad[:]
        
        # LBFGS save
        self.save_LBFGS(x, grad)
        
    
    def save_LBFGS(self, x, grad):
        ''' save values for LBFGS
        '''
        
        if(self.cpt_lbfgs <= self.l):
            # if the number of stored pairs does not exceed the maximum value, then save x and grad
            self.sk[:, self.cpt_lbfgs-1] = x[:]
            self.yk[:, self.cpt_lbfgs-1] = grad[:]
        else:
            # otherwise, erase the oldest pair and save the new one (shift)                         
            for i in range (self.l-1):
                self.sk[:, i] = self.sk[:, i+1]
                self.yk[:, i] = self.yk[:, i+1]
        
            self.sk[:, self.l-1] = x[:]
            self.yk[:, self.l-1] = grad[:]
        

    def update_LBFGS(self, x, grad):
        ''' update values for LBFGS
        '''

        if(self.cpt_lbfgs <= self.l):
            # if the number of stored pairs does not exceed the maximum value, 
            # then compute a new pair sk yk and update the counter cpt_lbfgs
            ii = self.cpt_lbfgs - 1
            self.sk[:, ii] = x    - self.sk[:, ii]
            self.yk[:, ii] = grad - self.yk[:, ii]
            self.cpt_lbfgs += 1
        else:
            # otherwise, simply update the lth pair
            self.sk[:,self.l-1] = x[:]    - self.sk[:,self.l-1]
            self.yk[:,self.l-1] = grad[:] - self.yk[:,self.l-1]


    def descent_LBFGS(self, grad):
        ''' Descent direction
        '''
        
        n = np.size(grad)
        borne_i = self.cpt_lbfgs - 1

        # Safeguard
        norml2_sk = normL2(self.sk[:,borne_i-1])
        norml2_yk = normL2(self.yk[:,borne_i-1])
        if( (norml2_sk == 0.) or (norml2_yk == 0.)):
            self.descent = -1. * grad
        else:
            # First phase of the recursion loop
            alpha = np.zeros(self.cpt_lbfgs)
            rho   = np.zeros(self.cpt_lbfgs)
            q     = np.zeros(n)
            q[:]  = grad[:]

            for i in range(borne_i):
                ii = borne_i - i - 1
                rho[ii] = 1. / scalL2(self.yk[:, ii], self.sk[:,ii])
                alpha[ii] = rho[ii] * scalL2(self.sk[:,ii], q)
                q[:] = q[:] - alpha[ii] * self.yk[:,ii]
        
            gamma_num = scalL2(self.sk[:,borne_i-1], self.yk[:,borne_i-1])
            gamma_den = normL2(self.yk[:,borne_i-1])

            # Scaling by gamma
            gamma = gamma_num / (gamma_den * gamma_den)
            self.descent[:] = gamma * q[:]

            # Second phase of the recursion loop
            for i in range (borne_i):
                beta = rho[i] * scalL2(self.yk[:,i], self.descent)
                self.descent[:] += (alpha[i] - beta) * self.sk[:, i]

            self.descent *= -1.


    #--------------------------------------------------------------------#
    #        Quasi-Newton preconditioned l-BFGS method: PLBFGS           # 
    #--------------------------------------------------------------------#
    def PLBFGS(self, x, fcost, grad, grad_preco):
        ''' Quasi-Newton preconditioned l-BFGS method
        '''

        if(self.FLAG == 'INIT'):
            # initialize the linesearch process
            self.init_PLBFGS(x, fcost, grad, grad_preco)
            x = self.linesearch(x, fcost, grad)
            self.print_info(fcost)
            self.FLAG = 'GRAD'
            self.nfwd_pb += 1

        elif(self.FLAG == 'PREC'):
            # if self.FLAG is PREC, we return from a call to the user preconditioner,
            # we have to finish the computation of the descent direction
            self.descent2_PLBFGS()

            # LBFGS save
            self.save_LBFGS(x, grad)
            self.cpt_iter += 1
            
            # before continuing we test for convergence
            if(self.std_test_conv(fcost)):
                self.FLAG = 'CONV'
                # print info on current iteration
                self.grad[:] = grad[:]
                self.print_info(fcost)
            else:
                self.FLAG = 'NSTE'
                # print info on current iteration
                self.grad[:] = grad[:]
                self.print_info(fcost)
        else:
            # else call the linesearch process
            x = self.linesearch(x, fcost, grad)
            if(self.task == 'NEW_STEP'): 
                # LBFGS update
                self.update_LBFGS(x, grad)
                # Start the computation of the new descent direction
                self.descent1_PLBFGS(grad)
                # Set self.FLAG to PREC for asking user to perform preconditioning 
                self.FLAG = 'PREC'                        
            elif(self.task == 'NEW_GRAD'): 
                # if the linesearch needs a new gradient: ask the user to provide it
                self.FLAG = 'GRAD'
                self.nfwd_pb += 1
            elif(self.task == 'FAILURE!'):
                # if the linesearch has failed, inform the user
                self.FLAG = 'FAIL'
                # print info on current iteration
                self.grad[:] = grad[:]
                self.print_info(fcost)
        return x

    def init_PLBFGS(self, x, fcost, grad, grad_preco):
        '''  Initilize linesearch parameters
        '''

        # set counters
        self.cpt_iter = 0
        self.f0 = fcost
        self.nfwd_pb = 0
        self.cpt_lbfgs = 1
        
        # initialize linesearch parameters
        self.fk = fcost

        # memory allocations
        n = np.size(x)

        self.xk = np.zeros_like(x)
        self.grad = np.zeros_like(x)
        self.descent = np.zeros_like(x)
        self.sk = np.zeros((n, self.l))
        self.yk = np.zeros((n, self.l))
        self.q_plb = np.zeros_like(x)

        self.xk[:] = x[:]
        self.grad[:] = grad[:]
        
        # first descent direction
        self.descent[:] = -1. * grad_preco[:]
        self.save_LBFGS(x, grad)


    # First loop
    def descent1_PLBFGS(self, grad):
        ''' First loop for PLBFGS
        '''

        borne_i = self.cpt_lbfgs - 1
        self.alpha_plb = np.zeros(self.cpt_lbfgs)
        self.rho_plb = np.zeros(self.cpt_lbfgs) 
        self.q_plb[:] = grad[:]

        for i in range(borne_i):
            ii = borne_i-i-1
            self.rho_plb[ii]   = 1. / scalL2(self.yk[:, ii], self.sk[:, ii])
            self.alpha_plb[ii] = self.rho_plb[ii] * scalL2(self.sk[:, ii], self.q_plb[:]) 
            self.q_plb[:]     -= self.alpha_plb[ii] * self.yk[:,ii]

    # Second loop
    def descent2_PLBFGS(self):
        ''' Second loop for PLBFGS
        '''
    
        borne_i = self.cpt_lbfgs-1  
        gamma_num = scalL2(self.sk[:, borne_i-1], self.yk[:, borne_i-1])
        gamma_den = normL2(self.yk[:, borne_i-1])
        gamma = gamma_num / (gamma_den**2)
        self.descent[:] = gamma * self.q_plb[:]
        for i in range(borne_i):
            beta = self.rho_plb[i] * scalL2(self.yk[:, i], self.descent[:])
            self.descent[:] += (self.alpha_plb[i]-beta) * self.sk[:,i]
        self.descent[:] *= -1.


    #--------------------------------------------------------------------#
    #                    Truncated Newton method: TRN                    # 
    #--------------------------------------------------------------------#
    def TRN(self, x, fcost, grad):
        ''' Truncated Newton method
        '''
    
        if(self.FLAG=='INIT'):
            # if FLAG is INIT, call the dedicated initialization def to allocate data structure self
            self.init_TRN( x, fcost, grad)
            self.print_info_TRN(fcost)
            self.comm = 'DESC'
            self.CG_phase = 'INIT'
            self.nfwd_pb = self.nfwd_pb+1
            self.conv_CG = False
            self.FLAG = 'NONE'

        if(self.comm == 'DESC'):
            # if self.comm is DES, the selfizer is computing a descent direction through the conjugate gradient
            self.descent_TRN(grad)
            
            if(self.conv_CG):
                # if the conjugate gradient has converged go to next phase: linesearch in the descent direction
                self.comm = 'NSTE'
                self.CG_phase = 'INIT'
                self.FLAG = 'NONE'
            else:
                # else perform a new iteration of conjugate gradient and ask the user to compute a Hessian-vector product
                self.FLAG = 'HESS'
                self.nhess = self.nhess+1
            
        elif(self.comm == 'NSTE'):
            # if self.comm is NSTE, the selfizer is looking for a new step in the descent direction
            x = self.linesearch(x, fcost, grad)
            if(self.task=='NEW_STEP'):          
                # if self.task is 'NEW_STEP, the linesearch process has found the new step
                self.cpt_iter += 1
                # Save the previous gradient norm 
                self.norm_grad_m1=self.norm_grad
                # Compute the new gradient norm 
                self.norm_grad = normL2(grad)
                # Print info on current nonlinear iteration
                self.print_info_TRN(fcost)
                # Test for convergence
                if(self.std_test_conv(fcost)):
                    self.FLAG = 'CONV'
                    self.print_info_TRN(fcost)
                else:
                    # Flags for the computation of the new descent direction
                    self.FLAG = 'NSTE'
                    self.comm = 'DESC'
                    # Update forcing term self.eta following the Eisenstat and Walker formula
                    self.forcing_term_TRN(grad)

            elif(self.task == 'NEW_GRAD'): # STILL SEARCHING THE STEP 
                # if self.task is 'NEW_GRAD, the linesearch process is continuing, the gradient at the current point is required
                self.FLAG = 'GRAD'
                self.nfwd_pb += 1

            elif(self.task =='FAILURE!'):
                # if self.task is 'FAILURE, the linesearch process has failed, the iterations are stopped
                self.FLAG = 'FAIL'
                self.print_info_TRN(fcost)

        return x

    def init_TRN(self, x, fcost, grad):
        '''  Initilize linesearch parameters
        '''
        
        # set counters
        self.cpt_iter = 0
        self.f0 = fcost   
        self.nfwd_pb = 0    
        self.nhess = 0
        self.eta = 0.9   
        #self.eta=0.1   
        #self.eta=1e-6   
            
        # initialize linesearch parameters
        self.fk = fcost
        self.cpt_iter_CG = 0
        
        # memory allocations
        self.xk = np.zeros_like(x)
        self.grad = np.zeros_like(x)
        self.descent = np.zeros_like(x)
        self.descent_prev = np.zeros_like(x)
        self.residual = np.zeros_like(x)
        self.d = np.zeros_like(x)
        self.Hd = np.zeros_like(x)
        self.eisenvect = np.zeros_like(x)

        self.xk[:] = x[:]
        self.grad[:] = grad[:]

        # norm of the first gradient
        self.norm_grad = normL2(grad)


    def descent_TRN(self, grad):
        ''' Descent direction
        '''

        if(self.CG_phase=='INIT'):
            # if self.CG_phase is INIT, initialize the conjugate gradient process
            self.residual[:] = grad[:]
            self.d[:] = -1.*self.residual[:]
            self.Hd[:] = 0.
            self.descent[:] = 0.
            self.qk_CG = 0.
            self.hessian_term = 0.
            self.norm_residual = normL2(self.residual)
            self.conv_CG = False
            self.cpt_iter_CG = 0
            self.print_info_TRN(0.0)
            self.CG_phase='IRUN'
        else:
            # else perform one conjugate gradient iteration
            dHd = scalL2(self.d, self.Hd)
            if(dHd < 0.):
                # if dHd < 0, detection of a negative eigenvalue of the Hessian operator, stop the process
                self.conv_CG = True
                print('Negative curvature')
                if(self.cpt_iter_CG == 0):
                    # if this is the first iteration, :return the opposite of the gradient as descent direction
                    self.descent[:] = self.d[:]

                    #-----------------------------------------------------#
                    # if the debug option is activated, compute the       #
                    # quadratic function minimized during the conjugate   #
                    # gradient process (check is this value decresae      #
                    # throughout the CG iterations )                      #
                    #-----------------------------------------------------#     
                    if(self.debug):
                        mgrad = np.zeros_like(grad)
                        mgrad[:] = -1.*grad[:]
                        self.norm_residual = normL2(self.residual)
                        alpha = (self.norm_residual**2) / dHd
                        self.qkm1_CG[:] = self.qk_CG[:]
                        grad_term = scalL2(self.descent,mgrad)
                        self.hessian_term = self.hessian_term+(alpha**2)*dHd
                        self.qk_CG = -grad_term + 0.5*self.hessian_term
            else:
                # if dHd > 0, : perform one conjugate gradient iteration
            
                # Update descent direction
                self.norm_residual = normL2(self.residual)
                alpha = (self.norm_residual**2)/dHd
                self.descent_prev[:] = self.descent[:]
                self.descent[:] = self.descent[:] + alpha*self.d[:]
                self.residual[:] = self.residual[:] + alpha*self.Hd[:]
            
                # Update CG direction
                norm_residual_prev = self.norm_residual
                self.norm_residual = normL2(self.residual)
                beta = (self.norm_residual**2)/(norm_residual_prev**2)
                self.d[:] = -1.*self.residual[:]+beta*self.d[:]
            
                # Update iteration counter 
                self.cpt_iter_CG = self.cpt_iter_CG+1
                #-----------------------------------------------------#
                # if the debug option is activated, compute the       #
                # quadratic function minimized during the conjugate   #
                # gradient process (check is this value decresae      #
                # throughout the CG iterations )                      #
                #-----------------------------------------------------#        
                if(self.debug):
                    mgrad = np.zeros_like(grad)
                    mgrad[:] = -1.*grad[:]
                    self.qkm1_CG = self.qk_CG
                    grad_term = scalL2(self.descent,mgrad)
                    descent_scal_Hd = scalL2(self.descent_prev,self.Hd)
                    self.hessian_term = self.hessian_term + (alpha**2) * dHd + 2. * alpha * descent_scal_Hd
                    self.qk_CG = -grad_term + 0.5*self.hessian_term
                
                # Check if the Eisenstat stopping critertion is satisfied
                self.conv_CG=( (self.norm_residual<=(self.eta*self.norm_grad)) or 
                    (self.cpt_iter_CG >= self.niter_max_CG))

                # Print information on the current CG iteration
                self.print_info_TRN(0.)


    def forcing_term_TRN(self, grad):
        ''' forcing_term_TRN
        '''
    
        # Computation of the forcing term self.eta following the formula
        eta_save = self.eta
        self.eisenvect[:] = grad[:] - self.residual[:]
        norm_eisenvect = normL2(self.eisenvect)
        self.eta = norm_eisenvect / self.norm_grad_m1
        
        # Additional safeguard if self.eta is too large
        eta_save_power = eta_save**((1.+np.sqrt(5.))/2.)
        if(eta_save_power > 0.1):
            self.eta = max(self.eta,eta_save_power)
        
        if(self.eta > 1.):
            self.eta = 0.9


    #--------------------------------------------------------------------#
    #         Preconditioned truncated Newton method: PTRN               # 
    #--------------------------------------------------------------------#
    def PTRN(self, x, fcost, grad, grad_preco):
        ''' Preconditioned truncated Newton method
        '''
    
        if(self.FLAG=='INIT'):
            # if FLAG is INIT, call the dedicated initialization def to allocate data structure self

            self.init_PTRN(x, fcost, grad)
            self.print_info_PTRN(fcost)
            self.comm = 'DES1'
            self.CG_phase = 'INIT'
            self.nfwd_pb = self.nfwd_pb+1
            self.conv_CG = False
            self.FLAG = 'NONE'
        
        if(self.comm == 'DES1'):
            # if self.comm is DES1, the selfizer starts the computation of a descent direction through the conjugate gradient
            if(self.CG_phase=='INIT'):
                # if self.CG_phase is INIT, initialization of the conjugate gradient process
                self.descent_PTRN0(grad, grad_preco)
                self.CG_phase = 'IRUN'
                self.comm = 'DES1'
                self.FLAG = 'HESS'
                self.nhess += 1
            elif(self.CG_phase == 'IRUN'):
                # if self.CG_phase is IRUN, iterate the conjugate gradient process (first part)
                
                self.descent_PTRN1(grad)
                if(self.conv_CG): 
                    #-----------------------------------------------------#
                    # if the conjugate gradient has already converged     # 
                    # (detection of a negative curvature), go to next     #
                    # phase: linesearch in the descent direction          #
                    #-----------------------------------------------------#
                    self.comm = 'NSTE'
                    self.CG_phase = 'INIT'        
                    self.FLAG = 'NONE'                   
                else:
                    # if the conjugate gradient has not converged prematurly,: ask the user to apply the preconditioner
                    self.FLAG = 'PREC'
                    self.comm = 'DES2'
                    
        elif(self.comm == 'DES2'):
            # if self.comm is DES2, the selfizer finish the current iteration of the conjugate gradient process
            self.descent_PTRN2(grad)

            if(self.conv_CG):
                # if the conjugate gradient has converged go to next phase: linesearch in the descent direction
                self.comm = 'NSTE'
                self.CG_phase = 'INIT'        
                self.FLAG = 'NONE'                
            else:
                # else start a new iteration of conjugate gradient and ask the user to compute a Hessian-vector product
                self.comm = 'DES1'
                self.FLAG = 'HESS'
                self.nhess = self.nhess + 1

        elif(self.comm == 'NSTE'):
            # if self.comm is NSTE, a descent direction has been computed, and a linesearch must be performed in this direction
            x = self.linesearch(x, fcost, grad)

            if(self.task=='NEW_STEP'):
                # if self.task is 'NEW_STEP, the linesearch process has found the new step
                self.cpt_iter += 1
                # Save the previous gradient norm 
                self.norm_grad_m1 = self.norm_grad
                # Computation of the new gradient norm 
                self.norm_grad = normL2(grad)
                # Print infor on current nonlinear iteration
                self.print_info_PTRN(fcost)        
                # Test for convergence
                if(self.std_test_conv(fcost)):
                    self.FLAG = 'CONV'
                    self.print_info_PTRN(fcost)
                else:
                    self.FLAG = 'NSTE'
                    # Flag for the computation of the new descent direction
                    self.comm = 'DES1'
                    # Update forcing term self.eta following the Eisenstat and Walker formula
                    self.forcing_term_TRN(grad)
                    
            elif(self.task == 'NEW_GRAD'): 
                # if self.task is 'NEW_GRAD, the linesearch process is continuing, the gradient at the current point is required
                self.FLAG = 'GRAD'         
                self.nfwd_pb += 1
            elif(self.task == 'FAILURE!'):        
                # if self.task is 'FAILURE, the linesearch process has failed, the iterations are stopped
                self.FLAG = 'FAIL'
                self.print_info_PTRN(fcost)

        return x

    def init_PTRN(self, x, fcost, grad):
        '''  Initilize linesearch parameters
        '''
        
        # set counters
        self.cpt_iter=0
        self.f0=fcost   
        self.nfwd_pb=0    
        self.nhess=0
        self.eta=0.9   
        #self.eta=0.1   
        #self.eta=1e-6   
            
        # initialize linesearch parameters
        self.fk=fcost
        self.cpt_iter_CG = 0
        
        # memory allocations
        self.xk = np.zeros_like(x)
        self.grad = np.zeros_like(x)
        self.descent = np.zeros_like(x)
        self.descent_prev = np.zeros_like(x)
        self.residual = np.zeros_like(x)
        self.residual_preco = np.zeros_like(x)
        self.d = np.zeros_like(x)
        self.Hd = np.zeros_like(x)
        self.eisenvect = np.zeros_like(x)

        self.xk[:] = x[:]
        self.grad[:] = grad[:]

        # norm of the first gradient
        self.norm_grad = normL2(grad)


    def descent_PTRN0(self, grad, grad_preco):
        ''' Descent direction
        '''
    
        # Initialization of the conjugate gradient process
        self.residual[:] = grad[:]
        self.residual_preco[:] = grad_preco[:]
        self.d[:] = -1. * self.residual_preco[:]
        self.Hd[:] = 0.
        self.descent[:] = 0.
        self.qk_CG = 0.
        self.hessian_term = 0.
        self.norm_residual = normL2(self.residual)
        self.conv_CG = False
        self.cpt_iter_CG = 0
        self.print_info_PTRN(0.0)

    def descent_PTRN1(self, grad):
        ''' descent_PTRN1
        '''
    
        # start one conjugate gradient iteration
        self.dHd = scalL2(self.d, self.Hd)
        if(self.dHd < 0.):
            # if dHd < 0, detection of a negative eigenvalue of the Hessian operator, stop the process
            self.conv_CG = True
            print('Negative curvature')
            if(self.cpt_iter_CG == 0):
                #-----------------------------------------------------#
                # if this is the first iteration,:return the          #
                # opposite of the preconditioned gradient as descent  #
                # direction: preconditioned steepest descent direction#
                #-----------------------------------------------------#   
                self.descent[:] = self.d[:]
                #-----------------------------------------------------#
                # if the debug option is activated, compute the       #
                # quadratic function minimized during the conjugate   #
                # gradient process (check is this value decresae      #
                # throughout the CG iterations )                      #
                #-----------------------------------------------------#
                if(self.debug):
                    self.res_scal_respreco = scalL2(self.residual,self.residual_preco)
                    self.alpha_CG = self.res_scal_respreco / self.dHd
                    self.qkm1_CG = self.qk_CG
                    mgrad = np.zeros_like(grad)
                    mgrad[:] = -1. * grad[:]
                    grad_term = scalL2(self.descent,mgrad)
                    self.hessian_term = self.hessian_term + (self.alpha_CG**2) * self.dHd
                    self.qk_CG = -grad_term + 0.5 * self.hessian_term
        else:
            # if dHd > 0,: start one conjugate gradient iteration

            # Update descent direction
            self.res_scal_respreco = scalL2(self.residual,self.residual_preco)
            self.alpha_CG = self.res_scal_respreco / self.dHd
            self.descent_prev[:] = self.descent[:]
            self.descent[:] = self.descent[:] + self.alpha_CG * self.d[:]
            self.residual[:] = self.residual[:] + self.alpha_CG * self.Hd[:]
            
            # STOP HERE and wait for preconditioning
        

    def descent_PTRN2(self, grad):
        ''' descent_PTRN2
        '''
        
        # continue the current conjugate gradient iteration
        # Update CG direction
        res_scal_respreco_prev = self.res_scal_respreco
        self.res_scal_respreco = scalL2(self.residual, self.residual_preco)
        beta = (self.res_scal_respreco) / (res_scal_respreco_prev)
        self.d[:] = -1. * self.residual_preco[:] + beta * self.d[:]
        
        # Update iteration counter 
        self.cpt_iter_CG += 1

        #-----------------------------------------------------#
        # if the debug option is activated, compute the       #
        # quadratic function minimized during the conjugate   #
        # gradient process (check is this value decresae      #
        # throughout the CG iterations )                      #
        #-----------------------------------------------------#  
        if(self.debug):
            self.qkm1_CG = self.qk_CG
            mgrad = np.zeros_like(grad)
            mgrad[:]=-1.*grad[:]
            grad_term = scalL2(self.descent,mgrad)
            descent_scal_Hd = scalL2(self.descent_prev,self.Hd)
            self.hessian_term = self.hessian_term+(self.alpha_CG**2) * self.dHd + 2. * self.alpha_CG * descent_scal_Hd
            self.qk_CG = -grad_term + 0.5 * self.hessian_term

        # Check if the Eisenstat stopping critertion is satisfied
        self.norm_residual = normL2(self.residual)
        self.conv_CG=((self.norm_residual<=(self.eta*self.norm_grad)) or 
                    (self.cpt_iter_CG>=self.niter_max_CG))
        
        # Print information on the current CG iteration
        self.print_info_PTRN(0.)


    #--------------------------------------------------------------------#
    #                               linesearch                           # 
    #--------------------------------------------------------------------#
    def linesearch(self, x, fcost, grad):
        ''' Linesearch
        '''
        if(self.first_ls):
            # FIRST LINESEARCH: initialization step
            self.fk = fcost
            self.q0 = scalL2(grad, self.descent)

            # Set the search interval bounds to 0
            self.alpha_L = 0.
            self.alpha_R = 0.
            self.task = 'NEW_GRAD'
            self.first_ls = False
            self.xk[:] = x[:]
            x[:] = self.xk[:] + self.alpha * self.descent[:]

            # If bounds activated, project x into the feasible range
            if(self.bound == 1):
                x = self.project(x)
            self.cpt_ls=0

        elif (self.cpt_ls >= self.nls_max and fcost < self.fk):
            # if the number of linesearch iteration outreaches the maximum allowed
            # but a decrease of the misfit is produced then accept the steplength

            self.task = 'NEW_STEP'
            self.first_ls = True
            
            # Compute new x in the descent direction
            x[:] = self.xk[:] + self.alpha * self.descent[:]
            
            # If bounds activated, project x into the feasible range
            if(self.bound == 1):
                x = self.project(x)

        elif(self.cpt_ls >= self.nls_max):
            # if the number of linesearch iteration outreaches the maximum allowed
            # without decreasing the misfit then the linesearch has failed
            self.task='FAILURE!'
            
        else:
            # If not initialization step and number of linesearch iteration ok
            # then perform one linesearch iteration                     
            
            self.q = scalL2(grad, self.descent)

            if (fcost <= (self.fk+(self.m1*self.alpha*self.q0))) and (self.q >= (self.m2*self.q0)):
                #--------------------------------------------------------------------#
                # First test if the Wolfe conditions are satisfied with              #     
                # current steplength, if this is the case, linesearch                # 
                # ends here with success                                             #
                #--------------------------------------------------------------------#
                self.task = 'NEW_STEP'
                self.first_ls = True
                if(self.debug):
                    print('fcost :', fcost)
                    print('optimize.f0    :', self.f0)
                    print('optimize.fk    :', self.fk)
                    print('optimize.alpha :', self.alpha)
                    print('optimize.q     :', self.q)
                    print('optimize.q0    :', self.q0)
                    print('m1             :', self.m1)
                    print('cpt_ls is      : ', self.cpt_ls)

            elif (fcost > (self.fk + (self.m1 * self.alpha * self.q0))):
                #--------------------------------------------------------------------#
                # If the first condition is not satisfied then shrink the            #
                # search interval                                                    #
                #--------------------------------------------------------------------#
                if(self.debug):
                    print('failure 1')
                    print('fcost          :', fcost)
                    print('optimize.fk    :', self.fk)
                    print('optimize.alpha :', self.alpha)
                    print('optimize.q0    :', self.q0)
                    print('m1             :', self.m1)
                    print('cpt_ls is      : ', self.cpt_ls)
                
                self.alpha_R = self.alpha
                new_alpha = (self.alpha_L+self.alpha_R) / 2.
                self.alpha = new_alpha
                self.task = 'NEW_GRAD'
                self.cpt_ls = self.cpt_ls+1

            elif( (fcost <= (self.fk + (self.m1 * self.alpha * self.q0))) and (self.q < (self.m2 * self.q0))):
                #--------------------------------------------------------------------#
                # If the second condition is not satisfied then shrink the           #
                # search interval unless the right bound of the search interval      #
                # as not yet been defined                                            #
                #--------------------------------------------------------------------#
                if(self.debug):
                    print('failure 2')
                    print('fcost          :', fcost)
                    print('optimize.fk    :', self.fk)
                    print('optimize.alpha :', self.alpha)
                    print('optimize.q0    :', self.q0)
                    print('optimize.q     :', self.q)
                    print('m1             :', self.m1)
                    print('m2             :', self.m2)
                    print('cpt_ls is      : ', self.cpt_ls)

                self.alpha_L = self.alpha

                if(self.alpha_R != 0.):
                    new_alpha = (self.alpha_L + self.alpha_R) / 2.
                else:
                    new_alpha = self.mult_factor * self.alpha
                
                self.alpha = new_alpha
                self.task = 'NEW_GRAD'
                self.cpt_ls = self.cpt_ls + 1
            
            # Compute new x in the descent direction
            x[:] = self.xk[:] + self.alpha * self.descent[:]
            # If bounds activated, project x into the feasible range
            if(self.bound == 1):
                x = self.project(x)
        return x

    def project(self, x):
        '''
        project the vector x into the box defined by the bound constraints optim.lb and optim.ub 
        '''

        n = np.size(x)
        for i in range(n):
            if(x[i] > self.ub[i]):
                x[i] = self.ub[i] - self.threshold
            if(x[i] < self.lb[i]):
                x[i] = self.lb[i] + self.threshold
        return x

    def std_test_conv(self, fcost):
        '''
        This routine implements a simple convergence test based on the relative cost decrease. 
        If the current relative cost is lower than a value set by the user in optim.conv
        then test_conv is returned equal to True, otherwise, it is returned equal to False
        '''
    
        test_conv = ((fcost/self.f0 < self.conv) or (self.cpt_iter >= self.niter_max))

        return test_conv


    def print_info(self, fcost):
        ''' print  information
        '''

        ng = normL2(self.grad)

        if(self.FLAG == 'INIT'):
            f = open('iterate_%s.log'%(self.method), 'w')
            if(self.method == 'SD'):
                f.write('**********************************************************************\n')
                f.write('         STEEEPEST DESCENT ALGORITHM         \n')
                f.write('**********************************************************************\n')
            elif(self.method == 'CG'):
                f.write('**********************************************************************\n')
                f.write('         NONLINEAR CONJUGATE GRADIENT ALGORITHM         \n')
                f.write('**********************************************************************\n')
            elif(self.method == 'LBFGS'):
                f.write('**********************************************************************\n')
                f.write('             l-BFGS ALGORITHM                \n')
                f.write('**********************************************************************\n')
            elif(self.method == 'PLBFGS'):
                f.write('**********************************************************************\n')
                f.write('             PRECONDITIONED l-BFGS ALGORITHM                \n')
                f.write('**********************************************************************\n')

            f.write('     Convergence criterion  : %10.2e\n' %self.conv)
            f.write('     Niter_max              : %10d\n'   %self.niter_max)
            f.write('     Initial cost is        : %10.2e\n' %self.f0)
            f.write('     Initial norm_grad is   : %10.2e\n' %ng)
            f.write('**********************************************************************\n')
            f.write('   Niter      fk         ||gk||       fk/f0        alpha        nls      ngrad    \n')
            f.write('%6d %12.2e %12.2e %12.2e %12.2e %8d %8d\n'
            %(self.cpt_iter, fcost, ng, fcost/self.f0, self.alpha, self.cpt_ls, self.nfwd_pb))
            f.close()

        elif(self.FLAG == 'CONV'):
            f = open('iterate_%s.log'%(self.method), 'a')
            f.write('%6d %12.2e %12.2e %12.2e %12.2e %8d %8d\n'
            %(self.cpt_iter, fcost, ng, fcost/self.f0, self.alpha, self.cpt_ls, self.nfwd_pb))
            f.write('**********************************************************************\n')

            if(self.cpt_iter >= self.niter_max):
                f.write('  STOP: MAXIMUM NUMBER OF ITERATION REACHED    \n')
            else:
                f.write('  STOP: CONVERGENCE CRITERION SATISFIED        \n')
                f.write('**********************************************************************\n')
                f.close()

        elif(self.FLAG == 'FAIL'):
            f = open('iterate_%s.log'%(self.method), 'a')
            f.write('**********************************************************************\n')
            f.write('  STOP: LINESEARCH FAILURE    \n')
            f.write('**********************************************************************\n')
            f.close()
        else:
            f = open('iterate_%s.log'%(self.method), 'a')
            f.write('%6d %12.2e %12.2e %12.2e %12.2e %8d %8d\n'
            %(self.cpt_iter, fcost, ng, fcost/self.f0, self.alpha, self.cpt_ls, self.nfwd_pb))
            f.close()


    def print_info_TRN(self, fcost):
        
        if(self.FLAG=='INIT') :
            f = open('iterate_TRN.log', 'w')
            f.write('******************************************************************************************\n')
            f.write('                                 TRUNCATED NEWTON ALGORITHM                               \n')
            f.write('******************************************************************************************\n')
            f.write('     Convergence criterion  : %10.2e \n' %self.conv)
            f.write('     Niter_max              : %7d \n'    %self.niter_max)
            f.write('     Initial cost is        : %10.2e \n' %self.f0)
            f.write('     Initial norm_grad is   : %10.2e \n' %self.norm_grad)
            f.write('     Maximum CG iter        : %7d \n'    %self.niter_max_CG)
            f.write('******************************************************************************************\n')
            f.write('   Niter      fk         ||gk||       fk/f0          alpha      nls     nit_CG      eta       ngrad     nhess \n')
            f.write('%6d %12.2e %12.2e %12.2e %12.2e %8d %8d %12.2e %8d %8d \n'%(
            self.cpt_iter, fcost, self.norm_grad, fcost/self.f0, self.alpha, self.cpt_ls, self.cpt_iter_CG, self.eta, self.nfwd_pb, self.nhess))
            f.close()

            f = open('iterate_TRN_CG.log', 'w')
            f.write('******************************************************************************************\n')
            f.write('                                 TRUNCATED NEWTON ALGORITHM                               \n')
            f.write('                                      INNER CG HISTORY                                    \n')
            f.write('******************************************************************************************\n')
            f.write('     Convergence criterion  : %10.2e \n' %self.conv)
            f.write('     Niter_max              : %7d \n'    %self.niter_max)
            f.write('     Initial cost is        : %10.2e \n' %self.f0)
            f.write('     Initial norm_grad is   : %10.2e \n' %self.norm_grad)
            f.write('     Maximum CG iter        : %7d \n'    %self.niter_max_CG)
            f.write('******************************************************************************************\n')
            f.close()
        elif(self.FLAG=='CONV'):
            f = open('iterate_TRN.log', 'a')
            f.write('**********************************************************************\n')
            if(self.cpt_iter == self.niter_max):
                f.write('  STOP: MAXIMUM NUMBER OF ITERATION REACHED    \n')
            else:
                f.write('  STOP: CONVERGENCE CRITERION SATISFIED        \n')
            
            f.write('**********************************************************************\n')
            f.close()

        elif(self.FLAG=='FAIL'):
            f = open('iterate_TRN.log', 'a')
            f.write('**********************************************************************\n')
            f.write('  STOP: LINESEARCH FAILURE    \n')
            f.write('**********************************************************************\n')
            f.close()
        elif(self.comm=='DESC'):
            f = open('iterate_TRN_CG.log', 'a')
            if(self.CG_phase=='INIT'):
                f.write('-------------------------------------------------------------------------------------------------\n')
                f.write(' NONLINEAR ITERATION %4d  ETA IS : %12.2e \n'%(self.cpt_iter, self.eta))
                f.write('-------------------------------------------------------------------------------------------------\n')
                f.write('  Iter_CG        qk       norm_res      norm_res/||gk||  \n')
                f.write('%8d %12.2e %12.2e %12.2e \n'%(self.cpt_iter_CG, self.qk_CG, self.norm_residual, self.norm_residual/self.norm_grad))
            else:
                f.write('%8d %12.2e %12.2e %12.2e \n'%(self.cpt_iter_CG, self.qk_CG, self.norm_residual, self.norm_residual/self.norm_grad))
            f.close()            
        else:
            f = open('iterate_TRN.log', 'a')
            f.write('%6d %12.2e %12.2e %12.2e %12.2e %8d %8d %12.2e %8d %8d \n'%(
            self.cpt_iter, fcost, self.norm_grad, fcost/self.f0, self.alpha, self.cpt_ls, self.cpt_iter_CG, self.eta, self.nfwd_pb, self.nhess))
            f.close()


    def print_info_PTRN(optim, fcost, FLAG):
        ''' Print information
        '''

        if(FLAG=='INIT'):
            f = open('iterate_PTRN.log', 'w')
            f.write('******************************************************************************************\n')
            f.write('                      PRECONDITIONED TRUNCATED NEWTON ALGORITHM                           \n')
            f.write('******************************************************************************************\n')
            f.write('     Convergence criterion  : %10.2e \n'%optim.conv)
            f.write('     Niter_max              : %7d \n'%optim.niter_max)
            f.write('     Initial cost is        : %10.2e \n'%optim.f0)
            f.write('     Initial norm_grad is   : %10.2e \n'%optim.norm_grad)
            f.write('     Maximum CG iter        : %7d \n'%optim.niter_max_CG)
            f.write('******************************************************************************************\n')
            f.write('   Niter      fk         ||gk||       fk/f0          alpha      nls     nit_CG      eta       ngrad     nhess \n')
            f.write('%6d %12.2e %12.2e %12.2e %12.2e %8d %8d %12.2e %8d %8d \n'%(
            optim.cpt_iter, fcost, optim.norm_grad, fcost/optim.f0, optim.alpha, optim.cpt_ls, optim.cpt_iter_CG, optim.eta, optim.nfwd_pb, optim.nhess))
            f.close()

            f = open('iterate_PTRN_CG.log', 'w')
            f.write('******************************************************************************************\n')
            f.write('                        PRECONDITIONED TRUNCATED NEWTON ALGORITHM                         \n')
            f.write('                                      INNER CG HISTORY                                    \n')
            f.write('******************************************************************************************\n')
            f.write('     Convergence criterion  : %10.2e \n'%optim.conv)
            f.write('     Niter_max              : %7d \n'%optim.niter_max)
            f.write('     Initial cost is        : %10.2e \n'%optim.f0)
            f.write('     Initial norm_grad is   : %10.2e \n'%optim.norm_grad)
            f.write('     Maximum CG iter        : %7d \n'%optim.niter_max_CG)
            f.write('******************************************************************************************\n')
            f.close()

        elif(FLAG=='CONV'):
            f = open('iterate_PTRN.log', 'a')
            f.write('**********************************************************************\n')
            if(optim.cpt_iter==optim.niter_max):
                f.write('  STOP: MAXIMUM NUMBER OF ITERATION REACHED    \n')
            else:
                f.write('  STOP: CONVERGENCE CRITERION SATISFIED        \n')
            f.write('**********************************************************************\n')
            f.close()

        elif(FLAG=='FAIL'):
            f = open('iterate_PTRN.log', 'a')
            f.write('**********************************************************************\n')
            f.write('  STOP: LINESEARCH FAILURE    \n')
            f.write('**********************************************************************\n')
            f.close()     
        elif(optim.comm=='DES1'):
            f = open('iterate_PTRN_CG.log', 'a')
            f.write('-------------------------------------------------------------------------------------------------\n')
            f.write(' NONLINEAR ITERATION %4d  ETA IS : %12.2e \n'%(optim.cpt_iter, optim.eta))
            f.write('-------------------------------------------------------------------------------------------------\n')
            f.write('  Iter_CG        qk       norm_res      norm_res/||gk||  \n')
            f.write('%8d %12.2e %12.2e %12.2e \n'%(optim.cpt_iter_CG, optim.qk_CG, optim.norm_residual, optim.norm_residual/optim.norm_grad))
            f.close()     

        elif(optim.comm=='DES2'):
            f = open('iterate_PTRN_CG.log', 'a')
            f.write('%8d %12.2e %12.2e %12.2e \n'%(optim.cpt_iter_CG, optim.qk_CG, optim.norm_residual, optim.norm_residual/optim.norm_grad))
            f.close()     
        else:
            f = open('iterate_PTRN.log', 'a')
            f.write('%6d %12.2e %12.2e %12.2e %12.2e %8d %8d %12.2e %8d %8d \n'%(
            optim.cpt_iter, fcost, optim.norm_grad, fcost/optim.f0, optim.alpha, optim.cpt_ls, optim.cpt_iter_CG, optim.eta, optim.nfwd_pb, optim.nhess))
            f.close()


    def print_info_PTRN(self, fcost):
        ''' Print information
        '''
        
        if(self.FLAG=='INIT'):
            f = open('iterate_PTRN.log', 'w')
            f.write('******************************************************************************************\n')
            f.write('                      PRECONDITIONED TRUNCATED NEWTON ALGORITHM                           \n')
            f.write('******************************************************************************************\n')
            f.write('     Convergence criterion  : %10.2e \n'%self.conv)
            f.write('     Niter_max              : %7d \n'   %self.niter_max)
            f.write('     Initial cost is        : %10.2e \n'%self.f0)
            f.write('     Initial norm_grad is   : %10.2e \n'%self.norm_grad)
            f.write('     Maximum CG iter        : %7d \n'   %self.niter_max_CG)
            f.write('******************************************************************************************\n')
            f.write('   Niter      fk         ||gk||       fk/f0          alpha      nls     nit_CG      eta       ngrad     nhess \n')
            f.write('%6d %12.2e %12.2e %12.2e %12.2e %8d %8d %12.2e %8d %8d \n'%(
            self.cpt_iter, fcost, self.norm_grad, fcost/self.f0, self.alpha, self.cpt_ls, self.cpt_iter_CG, self.eta, self.nfwd_pb, self.nhess))
            f.close()

            f = open('iterate_PTRN_CG.log', 'w')
            f.write('******************************************************************************************\n')
            f.write('                        PRECONDITIONED TRUNCATED NEWTON ALGORITHM                         \n')
            f.write('                                      INNER CG HISTORY                                    \n')
            f.write('******************************************************************************************\n')
            f.write('     Convergence criterion  : %10.2e \n'%self.conv)
            f.write('     Niter_max              : %7d \n'   %self.niter_max)
            f.write('     Initial cost is        : %10.2e \n'%self.f0)
            f.write('     Initial norm_grad is   : %10.2e \n'%self.norm_grad)
            f.write('     Maximum CG iter        : %7d \n'   %self.niter_max_CG)
            f.write('******************************************************************************************\n')
            f.close()

        elif(self.FLAG=='CONV'):
            f = open('iterate_PTRN.log', 'a')
            f.write('**********************************************************************\n')
            if(self.cpt_iter == self.niter_max):
                f.write('  STOP: MAXIMUM NUMBER OF ITERATION REACHED    \n')
            else:
                f.write('  STOP: CONVERGENCE CRITERION SATISFIED        \n')
            f.write('**********************************************************************\n')
            f.close()

        elif(self.FLAG=='FAIL'):
            f = open('iterate_PTRN.log', 'a')
            f.write('**********************************************************************\n')
            f.write('  STOP: LINESEARCH FAILURE    \n')
            f.write('**********************************************************************\n')
            f.close()     
        elif(self.comm=='DES1'):
            f = open('iterate_PTRN_CG.log', 'a')
            f.write('-------------------------------------------------------------------------------------------------\n')
            f.write(' NONLINEAR ITERATION %4d  ETA IS : %12.2e \n'%(self.cpt_iter, self.eta))
            f.write('-------------------------------------------------------------------------------------------------\n')
            f.write('  Iter_CG        qk       norm_res      norm_res/||gk||  \n')
            f.write('%8d %12.2e %12.2e %12.2e \n'%(self.cpt_iter_CG, self.qk_CG, self.norm_residual, self.norm_residual/self.norm_grad))
            f.close()     

        elif(self.comm=='DES2'):
            f = open('iterate_PTRN_CG.log', 'a')
            f.write('%8d %12.2e %12.2e %12.2e \n'%(self.cpt_iter_CG, self.qk_CG, self.norm_residual, self.norm_residual/self.norm_grad))
            f.close()     
        else:
            f = open('iterate_PTRN.log', 'a')
            f.write('%6d %12.2e %12.2e %12.2e %12.2e %8d %8d %12.2e %8d %8d \n'%(
            self.cpt_iter, fcost, self.norm_grad, fcost/self.f0, self.alpha, self.cpt_ls, self.cpt_iter_CG, self.eta, self.nfwd_pb, self.nhess))
            f.close()

# =================================================================================================
# utility functions
# =================================================================================================

def normL2(x):
    '''The routine normL2 returns the Euclidian norm of a vector x of size n
    
    Args:

    Returns:
    '''

    return np.linalg.norm(x)


def scalL2(x, y):
    '''
    The routine scalL2 returns the scalar product between two vectors x and y
    '''

    return np.inner(x, y)


class Swit_Optimizer(Optimization):
    ''' Optimizer class describes the optimization algorithm

    Parameters
    ----------
        vp_init : 2D array (float32)
            initial velocity model
        method : str
            optimization method, 'SD', 'CG', 'LBFGS', 'PLBFGS'
        niter_max : int
            maximum number of iterations
        vp_max : float
            maximum velocity in the bound constraint
        vp_min : float
            minimum velocity in the bound constraint
        debug : bool
            whether to output the debug information
    '''

    def __init__(self, 
                vp_init     = None,
                method      = 'SD', 
                niter_max   = 20,
                bound       = False, 
                vp_max      = 5000, 
                vp_min      = 1000, 
                debug       = False, 
                ):
        ''' initialize the optimizer class
        '''

        # set up the bound constraint if bound is True
        lb = vp_min * np.ones_like(vp_init).flatten() if bound else None
        ub = vp_max * np.ones_like(vp_init).flatten() if bound else None

        # inherit the SEISCOPE Optimization class
        Optimization.__init__(self, niter_max = niter_max, conv = 1e-6, 
                             method = method, bound = bound, lb = lb, ub = ub, 
                             lbfgs_memory = 5, debug = debug)

        # check the optimizer parameters
        self.__check__()

    def __check__(self):
        ''' check the optimizer parameters
        '''

        # check the initial model
        if self.vp_init is None:
        # if self.vp_init is None or self.rho_init is None:
            raise ValueError('Optimizer: the initial vp and rho models and gradient mask should be provided')
    
        # check the optimization method
        methods = ['sd', 'pnlcg', 'lbfgs', 'plbfgs','lbfgs','adam']
        if self.method.lower() not in methods:
            msg = 'Optimizer: method must be one of {}'.format(methods)
            err = 'Optimizer: unknown simulation type: {}'.format(self.method)
            raise ValueError(msg + '\n' + err)

        # check the maximum number of iterations
        if self.niter_max <= 0:
            raise ValueError('Optimizer: the maximum number of iterations should be larger than 0')

        # check the bound constraint
        if self.bound:
            if self.vp_max <= self.vp_min:
                raise ValueError('Optimizer: the maximum velocity should be larger than the minimum velocity')
            
            if self.vp_max < self.vp_init.max() or self.vp_min > self.vp_init.min():
                msg  = '\nOptimizer: the range of initial model is {:.2f} ~ {:.2f} m/s\n'.format(self.vp_init.min(), self.vp_init.max())
                msg += 'Optimizer: the range of bound constraint is {:.2f} ~ {:.2f} m/s\n'.format(self.vp_min, self.vp_max)
                err = 'Optimizer: the bound constraint should be larger than the range of initial model\n'
                raise ValueError(msg + err)


import numpy as np
import torch

##########################################################################
#                          numpy <=====> tensor     
##########################################################################

def numpy2tensor(a,dtype=torch.float32):
    """
        transform numpy data into tensor
    """
    if not torch.is_tensor(a):
        return torch.tensor(a,requires_grad=False,dtype=dtype)
        # return torch.from_numpy(a)
    else:
        return a

def tensor2numpy(a):
    """
        transform tensor data into numpy
    """
    if not torch.is_tensor(a):
        return a 
    else:
        return a.detach().numpy()
    
def gpu2cpu(a):
    if torch.is_tensor(a):
        if a.requires_grad:
            if a.device == 'cpu':
                a = a.detach().numpy()
            else:
                a = a.cpu().detach().numpy()
        else:
            if a.device == 'cpu':
                return a.numpy()
            else:
                return a.cpu().numpy()
    return a

##########################################################################
#                          list <=====> numpy     
##########################################################################
def list2numpy(a):
    """
        transform numpy data into tensor
    """
    if isinstance(a,list):
        return np.array(a)
    else:
        return a

def numpy2list(a):
    """
        transform numpy data into tensor
    """
    if not isinstance(a,list):
        return a.tolist()
    else:
        return a


"""
* Modified from: https://github.com/seisfwi/SWIT
* Original Author: Haipeng Li
* Original Author Email: haipeng@stanford.edu
=========================================================
* Author: Liu Feng (SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-04-20 09:32:43
* LastEditors: Liu Feng
* LastEditTime: 2024-05-15 19:30:38
* Description: 
* Copyright (c) 2024 by Liu Feng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
"""

import numpy as np
from typing import List, Optional
import numpy as np
from ADFWI.utils import list2numpy,numpy2list
from ADFWI.view import plot_wavelet

class Source(object):
    """Seismic Source class
    """
    def __init__(self,nt:int,dt:float,f0:float)->None:
        self.nt             = nt
        self.dt             = dt
        self.f0             = f0
        self.t              = np.arange(nt)*dt
        self.loc_x          = []
        self.loc_z          = []
        self.loc            = []
        self.type           = []
        self.wavelet        = []
        self.moment_tensor  = []
        self.num            = 0
    

    def __repr__(self):
        """Reimplement the repr function for printing the source information"""
        try:
            src_x = list2numpy(self.loc_x)
            src_z = list2numpy(self.loc_z)
            xmin = src_x.min()
            xmax = src_x.max()
            zmin = src_z.min()
            zmax = src_z.max()

            info = f"Seismic Source:\n"
            info += f"  Source wavelet: {self.nt} samples at {self.dt * 1000:.2f} ms\n"
            info += f"  Source number : {self.num}\n"
            info += f"  Source types  : {self.get_type(unique = True)}\n"
            info += f"  Source x range: {xmin} - {xmax} (grids)\n"
            info += f"  Source z range: {zmin} - {zmax} (grids)\n"
        except:
            info = f"Seismic Source:\n"
            info += f"  empty\n"
        return info
    
    def add_encoded_sources(self,
            src_x       : np.array,
            src_z       : np.array,
            src_wavelet : np.ndarray, #3D: [encode_n,[src_n,t]]
            src_type    : Optional[str]='mt',
            src_mt      : Optional[np.ndarray] = np.array([[1,0,0],[0,1,0],[0,0,1]]), 
        ) -> None:
        """source encoded
        """
        if src_x.shape != src_z.shape:
            raise ValueError(
                "Source location along x and z direction must have the same shape"
            )
        if src_type.lower() not in ["mt"]:
            raise ValueError(
                "Source type must be either mt"
            )
        if src_wavelet.shape[-1] != self.nt:
            raise ValueError(
                "Source wavelet must have the same length as the number of time samples"
            )
        if src_mt.shape != (3, 3):
            raise ValueError("Moment tensor must be a 3x3 matrix")

        if src_type.lower() == "mt" and src_mt is None:
            raise ValueError("Moment tensor must be provided for mt source")
        src_n = len(src_x)
        # add source
        self.loc_x.extend(numpy2list(src_x))
        self.loc_z.extend(numpy2list(src_z))
        self.type.extend([src_type]*src_n)
        self.wavelet.extend(src_wavelet)
        self.moment_tensor.extend(np.ones((src_n,3,3))*src_mt)
        self.num += src_n
        return
    
    def add_sources(self,
            src_x       : np.array,
            src_z       : np.array,
            src_wavelet : np.ndarray,
            src_type    : Optional[str]='mt',
            src_mt      : Optional[np.ndarray] = np.array([[1,0,0],[0,1,0],[0,0,1]]), 
        ) -> None:
        """add multiple sources with same wavelet
        """
        if src_x.shape != src_z.shape:
            raise ValueError(
                "Source location along x and z direction must have the same shape"
            )
        if src_type.lower() not in ["mt"]:
            raise ValueError(
                "Source type must be either mt"
            )
        if src_wavelet.shape[0] != self.nt:
            raise ValueError(
                "Source wavelet must have the same length as the number of time samples"
            )
        if src_mt.shape != (3, 3):
            raise ValueError("Moment tensor must be a 3x3 matrix")

        if src_type.lower() == "mt" and src_mt is None:
            raise ValueError("Moment tensor must be provided for mt source")
        src_n = len(src_x)
        # add source
        self.loc_x.extend(numpy2list(src_x.reshape(-1)))
        self.loc_z.extend(numpy2list(src_z.reshape(-1)))
        self.type.extend([src_type]*src_n)
        self.wavelet.extend(np.ones((src_n,self.nt))*src_wavelet)
        self.moment_tensor.extend(np.ones((src_n,3,3))*src_mt)
        self.num += src_n
        return
        
    def add_source(self,
            src_x       : int,
            src_z       : int,
            src_wavelet : np.ndarray,
            src_type    : Optional[str]='mt',
            src_mt      : Optional[np.ndarray] = np.array([[1,0,0],[0,1,0],[0,0,1]]), 
        ) -> None:
        """Append single source
        """
        if src_type.lower() not in ["mt"]:
            raise ValueError(
                "Source type must be either mt"
            )
        if src_wavelet.shape[0] != self.nt:
            raise ValueError(
                "Source wavelet must have the same length as the number of time samples"
            )
        if src_mt.shape != (3, 3):
            raise ValueError("Moment tensor must be a 3x3 matrix")

        if src_type.lower() == "mt" and src_mt is None:
            raise ValueError("Moment tensor must be provided for mt source")

        # add source
        self.loc_x.append(src_x)
        self.loc_z.append(src_z)
        self.type.append(src_type)
        self.wavelet.append(src_wavelet)
        self.moment_tensor.append(src_mt)
        self.num += 1
    
    def get_loc(self):
        """Return the source location
        """
        src_x = list2numpy(self.loc_x)
        src_z = list2numpy(self.loc_z)
        if len(list2numpy(self.loc_x).shape) == 1:
            src_x   = src_x.reshape(-1,1)
            src_z   = src_z.reshape(-1, 1)
            src_loc = np.hstack((src_x,src_z))
        else:
            
            src_loc = np.concatenate((src_x[..., np.newaxis], 
                                      src_z[..., np.newaxis]), 
                                      axis=-1)  # Add new axis and concatenate
        self.loc = src_loc.copy()
        return src_loc 
    
    def get_wavelet(self):
        """Return the source wavelets
        """
        wavelet = list2numpy(self.wavelet)
        return wavelet
    
    def get_moment_tensor(self):
        """Return the source wavelets
        """
        mt = list2numpy(self.moment_tensor)
        return mt
    
    def get_type(self, unique=False) -> List[str]:
        """Return the source type
        """
        type = list2numpy(self.type)
        
        if unique:
            type = list2numpy(list(set(self.type)))
        return type
    
    def plot_wavelet(self,index=0,src_idx=None,**kwargs):
        tlist = self.t
        wavelet = self.get_wavelet()
        # for encoded source
        if len(wavelet.shape) == 3:
            wavelet = wavelet[index][src_idx]
        else:
            wavelet = wavelet[index]
        plot_wavelet(tlist,wavelet,**kwargs)


from .utils import (numpy2tensor,tensor2numpy,gpu2cpu,list2numpy,numpy2list)

from .wavelets import wavelet

from .velocityDemo import (build_layer_model,
                           get_smooth_layer_model,
                           load_marmousi_model,
                           get_linear_vel_model,
                           get_linear_marmousi2_model,
                           resample_marmousi_model,
                           get_smooth_marmousi_model,
                           load_overthrust_model,
                           load_overthrust_initial_model,
                           resample_overthrust_model,
                           build_anomaly_background_model,
                           get_anomaly_model,
                           get_linear_vel_model,
                           load_valhall_model,
                           get_smooth_valhall_model,
                           load_hess_model,
                           get_smooth_hess_model,
                           get_linear_hess_model)

from .frequency_domin_process import *


'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-05-05 19:51:52
* LastEditors: LiuFeng
* LastEditTime: 2024-05-05 20:29:15
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''

from abc import abstractmethod
import torch
import numpy as np
from typing import Optional


def regular_StepLR(iter,step_size,alpha,gamma=0.8):
    n = iter//step_size
    return alpha*np.power(gamma,n)

def _l1_norm(x):
    return torch.sum(torch.abs(x))

def _l2_norm(x,eps = 1e-9):
    return torch.sqrt(torch.sum(x*x)+eps)

class Regularization():
    def __init__(self,nx:int,nz:int,dx:float,dz:float,
                 alphax:float,alphaz:float,
                 step_size:Optional[int]=1000,gamma:Optional[int]=1,
                 device="cpu",dtype=torch.float32) -> None:
        """
        Parameters:
        -------------
            nx (int)        : Number of grid points in x-direction
            nz (int)        : Number of grid points in z-direction
            dx (float)      : Grid size in x-direction (m)
            dz (float)      : Grid size in z-direction (m)
            alphax (float)  : the regular factor in x-direction
            alphaz (float)  : the regular factor in z-direction
            step_size (int) : the update step for alphax and alphaz 
            gamma (float)   : the update step decay factor 
        """
        self.iter       = 0
        self.step_size  = step_size
        self.gamma      = gamma
        self.alphax     = alphax
        self.alphaz     = alphaz
        self.nx         = nx
        self.nz         = nz
        self.dx         = dx
        self.dz         = dz
        self.device     = device
        self.dtype      = dtype
    
    @abstractmethod
    def forward(self):
        pass



'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-04-20 09:32:43
* LastEditors: LiuFeng
* LastEditTime: 2024-05-15 19:30:38
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''

from .base import Misfit
import torch

class Misfit_waveform_studentT(Misfit):
    """ Roubust T Loss (Alvaro et al., 2023; Guo et al., 2023)
            https://arxiv.org/pdf/2306.00753.pdf
            https://watermark.silverchair.com/gxac096.pdf
    
    Paraemters:
    -----------
        s (float)       : the number of degree of freedom (usually 1 or 2). 
        sigma (float)   : a scale parameter
        obs (Tensors)   : the observed waveform 
        syn (Tensors)   : the synthetic waveform 
    """
    def __init__(self,s=1,sigma=1,dt=1) -> None:
        self.s      = s
        self.sigma  = sigma
        self.dt     = dt

    def forward(self,obs,syn):
        rsd     = syn - obs
        loss    = 0.5*(self.s + 1)*torch.log(1 + rsd**2/(self.s*self.sigma**2))
        loss    = torch.sum(loss*self.dt)
        return loss